import unittest

from xlcalculator.xlfunctions import text, xl, xlerrors, func_xltypes


class TextModuleTest(unittest.TestCase):

    def test_CONCAT(self):
        self.assertEqual(
            text.CONCAT("SPAM", " ", "SPAM"), 'SPAM SPAM')
        self.assertEqual(
            text.CONCAT(
                "SPAM",
                " ",
                func_xltypes.Array([[1, 2]]),
                4
            ),
            'SPAM 124'
        )

    def test_CONCAT_with_too_many_items(self):
        self.assertIsInstance(
            text.CONCAT(*[0] * 300),
            xlerrors.ValueExcelError
        )

    def test_CONCAT_MS_cases(self):

        self.assertEqual(
            text.CONCAT(
                "The",
                " ",
                "sun",
                " ",
                "will",
                " ",
                "come",
                " ",
                "up",
                " ",
                "tomorrow."
            ),
            "The sun will come up tomorrow.")

        range0 = func_xltypes.Array(
            ["A's", 'a1', 'a2', 'a4', 'a5', 'a6', 'a7'])
        range1 = func_xltypes.Array(
            ["B's", 'b1', 'b2', 'b4', 'b5', 'b6', 'b7'])

        self.assertEqual(text.CONCAT(range0, range1),
                         "A'sa1a2a4a5a6a7B'sb1b2b4b5b6b7")

    def test_EXACT(self):
        self.assertEqual(text.EXACT("word", "word"), True)
        self.assertEqual(text.EXACT("Word", "word"), False)
        self.assertEqual(text.EXACT("w ord", "word"), False)

    def test_FIND(self):
        self.assertEqual(text.FIND('M', 'Miriam McGovern'), 1)
        self.assertEqual(text.FIND('m', 'Miriam McGovern'), 6)
        self.assertEqual(text.FIND('M', 'Miriam McGovern', 3), 8)

    def test_FIND_ValueError(self):
        self.assertIsInstance(
            text.FIND('B', 'Miriam McGovern'), xlerrors.ValueExcelError)

    def test_LEFT(self):
        self.assertEqual(text.LEFT('Sale Price', 4), "Sale")
        self.assertEqual(text.LEFT('Sweden'), "S")

    def test_LEN(self):
        self.assertEqual(text.LEN("Phoenix, AZ"), 11)
        self.assertEqual(text.LEN(""), 0)
        self.assertEqual(text.LEN("        One"), 11)

    def test_LOWER(self):
        self.assertEqual(text.LOWER("E. E. Cummings"), "e. e. cummings")
        self.assertEqual(text.LOWER("Apt. 2B"), "apt. 2b")

    def test_MID(self):
        self.assertEqual(text.MID('Romain', 3, 4), 'main')
        self.assertEqual(text.MID('Romain', 1, 2), 'Ro')
        self.assertEqual(text.MID('Romain', 3, 6), 'main')

    def test_MID_start_num_converted_to_integer(self):
        self.assertEqual(text.MID('Romain', 1.1, 2), 'Ro')

    def test_MID_num_chars_converted_to_integer(self):
        self.assertEqual(text.MID('Romain', 1, 2.1), 'Ro')

    def test_MID_start_num_must_be_superior_or_equal_to_1(self):
        self.assertIsInstance(
            text.MID('Romain', 0, 3), xlerrors.NumExcelError)

    def test_MID_num_chars_must_be_positive(self):
        self.assertIsInstance(
            text.MID('Romain', 1, -1), xlerrors.NumExcelError)

    def test_MID_with_too_large_text(self):
        self.assertIsInstance(
            text.MID('foo' + ' ' * xl.CELL_CHARACTER_LIMIT, 1, 3),
            xlerrors.ValueExcelError)

    def test_REPLACE(self):
        self.assertEqual(text.REPLACE("abcdefghijk", 6, 5, "*"), "abcde*k")
        self.assertEqual(text.REPLACE("2009", 3, 2, "10"), "2010")
        self.assertEqual(text.REPLACE("123456", 1, 3, "@"), "@456")

    def test_RIGHT(self):
        self.assertEqual(text.RIGHT("Sale Price", 5), "Price")
        self.assertEqual(text.RIGHT("Stock Number"), "r")

    def test_TRIM(self):
        self.assertEqual(text.TRIM(" First Quarter Earnings "),
                         "First Quarter Earnings")

    def test_UPPER(self):
        self.assertEqual(text.UPPER("total"), "TOTAL")
        self.assertEqual(text.UPPER("Yield"), "YIELD")



import os
import unittest
from dataclasses import dataclass
from decimal import Decimal, ROUND_UP, ROUND_DOWN

from xlcalculator import model, evaluator


RESOURCE_DIR = os.path.join(os.path.dirname(__file__), 'resources')


def get_resource(filename):
    return os.path.join(RESOURCE_DIR, filename)


@dataclass
class f_token:

    tvalue: str
    ttype: str
    tsubtype: str

    @classmethod
    def from_token(cls, token):
        return cls(token.tvalue, token.ttype, token.tsubtype)

    def __repr__(self):
        return "<{} tvalue: {} ttype: {} tsubtype: {}>".format(
            self.__class__.__name__, self.tvalue, self.ttype, self.tsubtype)

    def __str__(self):
        return self.__repr__()


class XlCalculatorTestCase(unittest.TestCase):

    def assertEqualRounded(self, lhs, rhs, rounding_precision=None):

        if rounding_precision is None:
            lhs_split = str(lhs).split('.')
            rhs_split = str(rhs).split('.')

            if len(lhs_split) > 1:
                len_lhs_after_decimal = len(lhs_split[1])
            else:
                len_lhs_after_decimal = None

            if len(rhs_split) > 1:
                len_rhs_after_decimal = len(rhs_split[1])
            else:
                len_rhs_after_decimal = None

            if len_lhs_after_decimal is None or len_rhs_after_decimal is None:
                return self.assertEqual(round(lhs), round(rhs))

            rounding_precision = min(
                len_lhs_after_decimal, len_rhs_after_decimal)

        precision_mask = "{0:." + str(rounding_precision - 1) + "f}1"
        precision = precision_mask.format(0.0)

        if lhs > rhs:
            lhs_value = Decimal(lhs).quantize(Decimal(
                precision), rounding=ROUND_DOWN)
            rhs_value = Decimal(rhs).quantize(
                Decimal(precision), rounding=ROUND_UP)
        else:
            lhs_value = Decimal(rhs).quantize(
                Decimal(precision), rounding=ROUND_UP)
            rhs_value = Decimal(lhs).quantize(
                Decimal(precision), rounding=ROUND_DOWN)

        return self.assertEqual(lhs_value, rhs_value)

    def assertEqualTruncated(self, lhs, rhs, truncating_places=None):
        lhs_before_dec, lhs_after_dec = str(lhs).split('.')
        rhs_before_dec, rhs_after_dec = str(rhs).split('.')

        if truncating_places is None:
            truncating_places = min(
                len(str(lhs).split('.')[1]), len(str(rhs).split('.')[1]))

        if 'E' in lhs_after_dec:
            lhs_value = float('.'.join((lhs_before_dec, lhs_after_dec)))
        else:
            lhs_value = float('.'.join((
                lhs_before_dec, lhs_after_dec[0:truncating_places])))

        if 'E' in lhs_after_dec:
            rhs_value = float('.'.join((rhs_before_dec, rhs_after_dec)))
        else:
            rhs_value = float('.'.join((
                rhs_before_dec, rhs_after_dec[0:truncating_places])))

        return self.assertAlmostEqual(lhs_value, rhs_value, truncating_places)

    def assertASTNodesEqual(self, lhs, rhs):
        lhs = [f_token.from_token(t) for t in lhs]
        rhs = [f_token.from_token(t) for t in rhs]
        return self.assertEqual(lhs, rhs)


class FunctionalTestCase(XlCalculatorTestCase):

    filename = None

    def setUp(self):
        compiler = model.ModelCompiler()
        self.model = compiler.read_and_parse_archive(
            get_resource(self.filename))
        self.evaluator = evaluator.Evaluator(self.model)



import unittest
import math as pymath
import mock

from xlcalculator.xlfunctions import math, xlerrors, func_xltypes


class MathModuleTest(unittest.TestCase):

    def test_ABS(self):
        self.assertEqual(math.ABS(3), 3)
        self.assertEqual(math.ABS('-3.4'), 3.4)

    def test_ABS_with_bad_arg(self):
        self.assertIsInstance(math.ABS('bad'), xlerrors.ValueExcelError)

    def test_ACOS(self):
        self.assertAlmostEqual(math.ACOS(-0.5), 2.094395102)

    def test_ACOSH(self):
        self.assertAlmostEqual(math.ACOSH(1), 0)
        self.assertAlmostEqual(math.ACOSH(10), 2.9932228)

    def test_ACOSH_less_than_1(self):
        self.assertIsInstance(math.ACOSH(-1), xlerrors.NameExcelError)

    def test_ASIN(self):
        self.assertAlmostEqual(math.ASIN(-0.5), -0.523598776)

    def test_ASIN_out_of_bounds(self):
        self.assertIsInstance(math.ASIN(-2), xlerrors.NumExcelError)
        self.assertIsInstance(math.ASIN(-2), xlerrors.NumExcelError)

    def test_ASINH(self):
        self.assertAlmostEqual(math.ASINH(-2.5), -1.647231146)
        self.assertAlmostEqual(math.ASINH(10), 2.99822295)

    def test_ATAN(self):
        self.assertAlmostEqual(math.ATAN(1), 0.785398163)

    def test_ARTAN2(self):
        self.assertAlmostEqual(math.ATAN2(1, 1), 0.785398163)
        self.assertAlmostEqual(math.ATAN2(-1, -1), -2.35619449)

    def test_CEILING_number(self):
        self.assertEqual(math.CEILING(2.5, 1), 3)
        self.assertEqual(math.CEILING(-2.5, -2), -4)
        self.assertEqual(math.CEILING(-2.5, 2), -2)
        self.assertEqual(math.CEILING(1.5, 0.1), 1.5)
        self.assertEqual(math.CEILING(0.234, 0.01), 0.24)
        self.assertEqual(math.CEILING(0, -2), 0)
        self.assertEqual(math.CEILING(2, 0), 0)
        self.assertIsInstance(math.CEILING(2, -2), xlerrors.NumExcelError)

    def test_COS(self):
        self.assertAlmostEqual(math.COS(1.047), 0.5001711)

    def test_COSH(self):
        self.assertAlmostEqual(math.COSH(4), 27.3082328)

    def test_DEGREES(self):
        self.assertEqual(math.DEGREES(pymath.pi), 180)

    def test_EVEN(self):
        self.assertEqual(math.EVEN(1.5), 2)
        self.assertEqual(math.EVEN(1.5), 2)
        self.assertEqual(math.EVEN(3), 4)
        self.assertEqual(math.EVEN(2), 2)
        self.assertEqual(math.EVEN(-1), -2)

    def test_EXP(self):
        self.assertAlmostEqual(math.EXP(1), 2.71828183)
        self.assertAlmostEqual(math.EXP(2), 7.3890561)

    def test_FACT(self):
        self.assertEqual(math.FACT(5), 120)
        self.assertEqual(math.FACT(1.9), 1)
        self.assertEqual(math.FACT(0), 1)
        self.assertIsInstance(math.FACT(-1), xlerrors.NumExcelError)
        self.assertEqual(math.FACT(1), 1)

    def test_FACTDOUBLE(self):
        self.assertEqual(math.FACTDOUBLE(6), 48)
        self.assertEqual(math.FACTDOUBLE(7), 105)
        self.assertEqual(math.FACTDOUBLE(0), 1)
        self.assertIsInstance(math.FACTDOUBLE(-1), xlerrors.NumExcelError)
        self.assertEqual(math.FACTDOUBLE(1), 1)

    def test_FLOOR(self):
        self.assertEqual(math.FLOOR(3.7, 2), 2)
        self.assertEqual(math.FLOOR(-2.5, -2), -2)
        self.assertEqual(math.FLOOR(1.58, 0.1), 1.5)
        self.assertEqual(math.FLOOR(0.234, 0.01), 0.23)

    def test_FLOOR_number(self):
        self.assertEqual(math.FLOOR(0, -2), 0)

    def test_FLOOR_significance(self):
        self.assertIsInstance(math.FLOOR(2, 0), xlerrors.DivZeroExcelError)
        self.assertIsInstance(math.FLOOR(2.5, -2), xlerrors.NumExcelError)

    def test_FLOOR_errors(self):
        self.assertIsInstance(math.FLOOR(2.5, -2), xlerrors.NumExcelError)

        self.assertIsInstance(math.FLOOR("hello", -2),
                              xlerrors.ValueExcelError)
        self.assertIsInstance(math.FLOOR(2.5, "hello"),
                              xlerrors.ValueExcelError)
        self.assertIsInstance(math.FLOOR("hello", "hello"),
                              xlerrors.ValueExcelError)

    def test_INT(self):
        self.assertEqual(math.INT(8.9), 8)
        self.assertEqual(math.INT(-8.9), -9)

    def test_LN(self):
        self.assertEqual(math.LN(2.718281828459045), 1)

    def test_LN_with_bad_arg(self):
        self.assertIsInstance(math.LN('bad'), xlerrors.ValueExcelError)

    def test_LOG(self):
        self.assertEqual(math.LOG(10), 1)
        self.assertEqual(math.LOG(8, 2), 3)
        self.assertAlmostEqual(math.LOG(86, 2.7182818), 4.4543473)

    def test_LOG10(self):
        self.assertAlmostEqual(math.LOG10(86), 1.93449845)
        self.assertEqual(math.LOG10(10), 1)
        self.assertEqual(math.LOG10(100000), 5)
        self.assertEqual(math.LOG10(1E+5), 5)

    def test_MOD(self):
        self.assertEqual(math.MOD(1, 2), 1)

    def test_MOD_with_float(self):
        self.assertEqual(math.MOD(1.0, 2), 1)

    def test_MOD_with_string(self):
        self.assertEqual(math.MOD('1.0', 2), 1)

    def test_MOD_with_bad_arg(self):
        self.assertIsInstance(math.MOD('bad', 2), xlerrors.ValueExcelError)
        self.assertIsInstance(math.MOD(1, 'bad'), xlerrors.ValueExcelError)

    def test_PI(self):
        self.assertEqual(math.PI(), 3.141592653589793)

    def test_POWER(self):
        self.assertEqual(math.POWER(10, 2), 100)
        self.assertEqual(math.POWER('-1', 2), 1)
        self.assertEqual(math.POWER(2.5, 2), 6.25)

    def test_POWER_with_bad_arg(self):
        self.assertIsInstance(math.POWER('bad', 2), xlerrors.ValueExcelError)
        self.assertIsInstance(math.POWER(10, 'bad'), xlerrors.ValueExcelError)

    def test_RADIANS(self):
        self.assertAlmostEqual(math.RADIANS(270), 4.712389)

    def test_RAND(self):
        with mock.patch.object(math, 'rand', lambda: 0.5):
            self.assertEqual(
                math.RAND(), 0.5)

    def test_RANDBETWEEN(self):
        with mock.patch.object(math, 'rand', lambda: 0.5):
            self.assertEqual(
                math.RANDBETWEEN(5, 10), 7)

    def test_ROUND(self):
        self.assertEqual(math.ROUND(0.6), 1)
        self.assertEqual(math.ROUND(1.3), 1)
        self.assertEqual(math.ROUND(1.25, 1), 1.3)

    def test_ROUND_with_bad_arg(self):
        self.assertIsInstance(
            math.ROUND('bad'), xlerrors.ValueExcelError)
        self.assertIsInstance(
            math.ROUND(1.3, 'bad'), xlerrors.ValueExcelError)

    def test_ROUNDUP(self):
        self.assertEqual(math.ROUNDUP(0.6), 1)
        self.assertEqual(math.ROUNDUP(1.3), 2)
        self.assertEqual(math.ROUNDUP(1.24, 1), 1.3)

    def test_ROUNDUP_with_bad_arg(self):
        self.assertIsInstance(
            math.ROUNDUP('bad'), xlerrors.ValueExcelError)
        self.assertIsInstance(
            math.ROUNDUP(1.3, 'bad'), xlerrors.ValueExcelError)

    def test_ROUNDDOWN(self):
        self.assertEqual(math.ROUNDDOWN(0.6), 0)
        self.assertEqual(math.ROUNDDOWN(1.3), 1)
        self.assertEqual(math.ROUNDDOWN(1.26, 1), 1.2)

    def test_ROUNDDOWN_with_bad_arg(self):
        self.assertIsInstance(
            math.ROUNDDOWN('bad'), xlerrors.ValueExcelError)
        self.assertIsInstance(
            math.ROUNDDOWN(1.3, 'bad'), xlerrors.ValueExcelError)

    def test_SIGN(self):
        self.assertEqual(math.SIGN(10), 1)
        self.assertEqual(math.SIGN(4 - 4), 0)
        self.assertEqual(math.SIGN(-0.00001), -1)

    def test_SIN(self):
        self.assertAlmostEqual(math.SIN(math.PI()), 0.0)
        self.assertAlmostEqual(math.SIN(math.PI() / 2), 1.0)
        self.assertAlmostEqual(math.SIN(30 * math.PI() / 180), 0.5)
        self.assertAlmostEqual(math.SIN(math.RADIANS(30)), 0.5)

    def test_SQRT(self):
        self.assertEqual(math.SQRT(4), 2)
        self.assertEqual(math.SQRT(4.0), 2.0)

    def test_SQRT_with_neg_number(self):
        self.assertIsInstance(math.SQRT(-4), xlerrors.NumExcelError)

    def test_SQRT_with_bad_arg(self):
        self.assertIsInstance(math.SQRT('bad'), xlerrors.ValueExcelError)

    def test_SQRTPI(self):
        self.assertAlmostEqual(math.SQRTPI(1), 1.77245385)
        self.assertAlmostEqual(math.SQRTPI(2), 2.50662827)

    def test_SQRTPI_negative_number(self):
        self.assertIsInstance(math.SQRTPI(-2), xlerrors.NumExcelError)

    def test_SUM(self):
        self.assertEqual(math.SUM(func_xltypes.Array([[1, 2], [3, 4]])), 10)
        self.assertEqual(math.SUM(1, 2, 3, 4.0), 10.0)

    def test_SUM_with_nonnumbers_in_range(self):
        self.assertEqual(math.SUM(func_xltypes.Array([[1, 'bad'], [3, 4]])), 8)
        self.assertEqual(math.SUM(
            func_xltypes.Array([
                [func_xltypes.Number(1), func_xltypes.Text('N/A')],
                [func_xltypes.Number(3), func_xltypes.Number(4)]
            ])), 8)

    def test_SUM_with_bad_Arg(self):
        self.assertEqual(math.SUM('foo'), 0)

    def test_SUM_empty(self):
        self.assertEqual(math.SUM(), 0)

    def test_SUMIF(self):
        self.assertEqual(math.SUMIF([0, 1, 2], '>=1', [10, 20, 30]), 50)

    def test_SUMIF_text(self):
        self.assertEqual(math.SUMIF(['a', 'b', 'A'], '=a', [10, 20, 30]), 40)

    def test_SUMIF_invalid_criteria(self):
        self.assertIsInstance(
            math.SUMIF([0, 1, 2], [0, 1], [10, 20, 30]),
            xlerrors.ValueExcelError)

    def test_SUMIF_unspecified_sum_range(self):
        self.assertEqual(math.SUMIF([0, 1, 2, 3], ">=2"), 5)

    def test_SUMIF_with_invalid_sum_range(self):
        # In this case, "bad" is converted to a single item array, then
        # filtered to an array where the value is 0, so that the sum is always
        # 0.
        self.assertEqual(math.SUMIF([0, 1, 2, 3], ">=2", 'bad'), 0)

    def test_SUMIFS(self):
        self.assertEqual(
            math.SUMIFS([10, 20, 30, 40],
                        [0, 1, 2, 3],
                        ">=1",
                        ["a", "b", "a", "A"],
                        "a"),
            70
        )

    def test_SUMIFS_invalid_criteria(self):
        self.assertIsInstance(
            math.SUMIFS([10, 20, 30], [0, 1, 2], [0, 1], ["a", "b", "a"], "a"),
            xlerrors.ValueExcelError
        )

    def test_SUMIFS_unequal_array_lengths(self):
        self.assertEqual(
            math.SUMIFS(
                [10, 20, 30], [0, 1, 2], ">=1", ["a", "b", "a", 1], "a"
            ),
            0
        )

    def test_SUMPRODUCT(self):
        range1 = func_xltypes.Array([[1], [10], [3]])
        range2 = func_xltypes.Array([[3], [1], [2]])
        self.assertEqual(math.SUMPRODUCT(range1, range2), 19)

    def test_SUMPRODUCT_ranges_with_different_sizes(self):
        range1 = func_xltypes.Array([[1], [10], [3]])
        range2 = func_xltypes.Array([[3], [3], [1], [2]])
        self.assertIsInstance(
            math.SUMPRODUCT(range1, range2), xlerrors.ValueExcelError)

    def test_SUMPRODUCT_with_empty_frist_range(self):
        self.assertEqual(math.SUMPRODUCT(func_xltypes.Array([])), 0)

    def test_SUMPRODUCT_without_any_range(self):
        self.assertIsInstance(math.SUMPRODUCT(), xlerrors.NullExcelError)

    def test_SUMPRODUCT_ranges_with_errors(self):
        range1 = func_xltypes.Array(
            [[xlerrors.NumExcelError('err')], [10], [3]]
        )
        range2 = func_xltypes.Array([[3], [3], [1]])
        self.assertIsInstance(
            math.SUMPRODUCT(range1, range2), xlerrors.NaExcelError)

    def test_SUMPRODUCT_with_single_value(self):
        self.assertEqual(math.SUMPRODUCT(1), 1.0)

    def test_TAN(self):
        self.assertAlmostEqual(math.TAN(0.785), 0.99920399)
        self.assertAlmostEqual(math.TAN(45 * math.PI() / 180), 1)
        self.assertAlmostEqual(math.TAN(math.RADIANS(45)), 1)

    def test_TRUNC(self):
        self.assertEqual(math.TRUNC(0.6), 0)
        self.assertEqual(math.TRUNC(1.3), 1)
        self.assertEqual(math.TRUNC(1.26, 1), 1.2)

    def test_TRUNC_with_bad_arg(self):
        self.assertIsInstance(
            math.TRUNC('bad'), xlerrors.ValueExcelError)
        self.assertIsInstance(
            math.TRUNC(1.3, 'bad'), xlerrors.ValueExcelError)





from . import testing


class CrossSheetTest(testing.FunctionalTestCase):

    filename = "cross_sheet.xlsx"

    def test_reference(self):
        """
            Tests if a reference to a cell on another sheet (Sheet1),
            which refers to a cell on it's own sheet (Sheet2) is resolved
            properly as being on Sheet2

            Also validates if after that passing in only "A1" refers back to
            Sheet1
        """
        # This tests going back and forth between original Sheet1 and Sheet2
        # multiple times in different orders
        self.assertEqual(
            self.evaluator.evaluate('Sheet1!A4'),
            5
        )



from xlcalculator.tokenizer import f_token, ExcelParser

from . import testing


class ExcelParserTest(testing.XlCalculatorTestCase):

    def setUp(self):
        self.operand_range_00 = ['A1:B1']
        self.operand_range_01 = ['A1', 'B1']

    def parse(self, formula):
        parser = ExcelParser()
        return parser.parse(formula).items

    def test_cell_minus(self):
        self.assertASTNodesEqual(
            self.parse('-A1'),
            [
                f_token(tvalue='-', ttype='operator-prefix', tsubtype=''),
                f_token(tvalue='A1', ttype='operand', tsubtype='range')
            ]
        )

    def test_cells_minus(self):
        self.assertASTNodesEqual(
            self.parse('-A1,-B1'),
            [
                f_token(tvalue='-', ttype='operator-prefix', tsubtype=''),
                f_token(tvalue='A1', ttype='operand', tsubtype='range'),
                f_token(tvalue=',', ttype='operator-infix', tsubtype='union'),
                f_token(tvalue='-', ttype='operator-prefix', tsubtype=''),
                f_token(tvalue='B1', ttype='operand', tsubtype='range')
            ]
        )

    def test_cell_minus_with_sheet(self):
        self.assertASTNodesEqual(
            self.parse('-Sheet1!A1'),
            [
                f_token(tvalue='-', ttype='operator-prefix', tsubtype=''),
                f_token(tvalue='Sheet1!A1', ttype='operand', tsubtype='range')
            ]
        )

    def test_cell_address(self):
        self.assertASTNodesEqual(
            self.parse('A1'),
            [
                f_token(tvalue='A1', ttype='operand', tsubtype='range')
            ]
        )

    def test_cell_address_by_number(self):
        self.assertASTNodesEqual(
            self.parse('R[1]C[1]'),
            [
                f_token(tvalue='R[1]C[1]', ttype='operand', tsubtype='range')
            ]
        )

    def test_cell_address_with_sheet(self):
        self.assertASTNodesEqual(
            self.parse('Sheet1!A1'),
            [
                f_token(tvalue='Sheet1!A1', ttype='operand', tsubtype='range')
            ]
        )

    def test_cells_union(self):
        self.assertASTNodesEqual(
            self.parse('A1,B1'),
            [
                f_token(tvalue='A1', ttype='operand', tsubtype='range'),
                f_token(tvalue=',', ttype='operator-infix', tsubtype='union'),
                f_token(tvalue='B1', ttype='operand', tsubtype='range')
            ]
        )

    def test_cells_union_with_same_sheet(self):
        self.assertASTNodesEqual(
            self.parse('Sheet1!A1,Sheet1!B1'),
            [
                f_token(tvalue='Sheet1!A1', ttype='operand', tsubtype='range'),
                f_token(tvalue=',', ttype='operator-infix', tsubtype='union'),
                f_token(tvalue='Sheet1!B1', ttype='operand', tsubtype='range')
            ]
        )

    def test_cells_union_with_different_sheet(self):
        self.assertASTNodesEqual(
            self.parse('Sheet1!A1,Sheet2!B1'),
            [
                f_token(tvalue='Sheet1!A1', ttype='operand', tsubtype='range'),
                f_token(tvalue=',', ttype='operator-infix', tsubtype='union'),
                f_token(tvalue='Sheet2!B1', ttype='operand', tsubtype='range')
            ]
        )

    def test_cells_addition(self):
        self.assertASTNodesEqual(
            self.parse('A1+B1'),
            [
                f_token(tvalue='A1', ttype='operand', tsubtype='range'),
                f_token(tvalue='+', ttype='operator-infix', tsubtype='math'),
                f_token(tvalue='B1', ttype='operand', tsubtype='range')
            ]
        )

    def test_cells_addition_with_same_sheet(self):
        self.assertASTNodesEqual(
            self.parse('Sheet1!A1+Sheet1!B1'),
            [
                f_token(tvalue='Sheet1!A1', ttype='operand', tsubtype='range'),
                f_token(tvalue='+', ttype='operator-infix', tsubtype='math'),
                f_token(tvalue='Sheet1!B1', ttype='operand', tsubtype='range')
            ]
        )

    def test_cells_addition_with_different_sheet(self):
        self.assertASTNodesEqual(
            self.parse('Sheet1!A1+Sheet2!B1'),
            [
                f_token(tvalue='Sheet1!A1', ttype='operand', tsubtype='range'),
                f_token(tvalue='+', ttype='operator-infix', tsubtype='math'),
                f_token(tvalue='Sheet2!B1', ttype='operand', tsubtype='range')
            ]
        )

    def test_cells_is_equal(self):
        self.assertASTNodesEqual(
            self.parse('A1=B1'),
            [
                f_token(tvalue='A1', ttype='operand', tsubtype='range'),
                f_token(
                    tvalue='=', ttype='operator-infix', tsubtype='logical'),
                f_token(tvalue='B1', ttype='operand', tsubtype='range')
            ]
        )

    def test_cells_is_equal_with_sheet(self):
        self.assertASTNodesEqual(
            self.parse('Sheet1!A1=Sheet1!B1'),
            [
                f_token(tvalue='Sheet1!A1', ttype='operand', tsubtype='range'),
                f_token(
                    tvalue='=', ttype='operator-infix', tsubtype='logical'),
                f_token(tvalue='Sheet1!B1', ttype='operand', tsubtype='range')
            ]
        )

    def test_cells_function(self):
        self.assertASTNodesEqual(
            self.parse('SUM(A1,B1)'),
            [
                f_token(tvalue='SUM', ttype='function', tsubtype='start'),
                f_token(tvalue='A1', ttype='operand', tsubtype='range'),
                f_token(tvalue=',', ttype='argument', tsubtype=''),
                f_token(tvalue='B1', ttype='operand', tsubtype='range'),
                f_token(tvalue='', ttype='function', tsubtype='stop')
            ]
        )

    def test_cells_function_with_trailing_comma(self):
        self.assertASTNodesEqual(
            self.parse('SUM(A1,)'),
            [
                f_token(tvalue='SUM', ttype='function', tsubtype='start'),
                f_token(tvalue='A1', ttype='operand', tsubtype='range'),
                f_token(tvalue=',', ttype='argument', tsubtype=''),
                f_token(tvalue='', ttype='function', tsubtype='stop')
            ]
        )

    def test_cells_function_with_sheet(self):
        self.assertASTNodesEqual(
            self.parse('SUM(Sheet1!A1,Sheet1!B1)'),
            [
                f_token(tvalue='SUM', ttype='function', tsubtype='start'),
                f_token(tvalue='Sheet1!A1', ttype='operand', tsubtype='range'),
                f_token(tvalue=',', ttype='argument', tsubtype=''),
                f_token(tvalue='Sheet1!B1', ttype='operand', tsubtype='range'),
                f_token(tvalue='', ttype='function', tsubtype='stop')
            ]
        )

    def test_cells_function_and_operator(self):
        self.assertASTNodesEqual(
            self.parse('SUM(A1+B1,C1)'),
            [
                f_token(tvalue='SUM', ttype='function', tsubtype='start'),
                f_token(tvalue='A1', ttype='operand', tsubtype='range'),
                f_token(tvalue='+', ttype='operator-infix', tsubtype='math'),
                f_token(tvalue='B1', ttype='operand', tsubtype='range'),
                f_token(tvalue=',', ttype='argument', tsubtype=''),
                f_token(tvalue='C1', ttype='operand', tsubtype='range'),
                f_token(tvalue='', ttype='function', tsubtype='stop')
            ]
        )

    def test_range_address(self):
        self.assertASTNodesEqual(
            self.parse('A1:B1'),
            [
                f_token(tvalue='A1:B1', ttype='operand', tsubtype='range')
            ]
        )

    def test_range_address_with_sheet(self):
        self.assertASTNodesEqual(
            self.parse('Sh1!A1:B1'),
            [
                f_token(tvalue='Sh1!A1:B1', ttype='operand', tsubtype='range')
            ]
        )

    def test_range_address_with_same_sheets(self):
        self.assertASTNodesEqual(
            self.parse('Sh1!A1:Sh1!B1'),
            [
                f_token(tvalue='Sh1!A1:Sh1!B1',
                        ttype='operand', tsubtype='range')
            ]
        )

    def test_range_address_with_different_sheets(self):
        # THIS CAN'T HAPPEN IN REAL LIFE
        self.assertASTNodesEqual(
            self.parse('Sh1!A1:Sh2!B1'),
            [
                f_token(tvalue='Sh1!A1:Sh2!B1',
                        ttype='operand', tsubtype='range')
            ]
        )

    def test_range_address_union(self):
        self.assertASTNodesEqual(
            self.parse('A1:B1,A2:B2'),
            [
                f_token(tvalue='A1:B1', ttype='operand', tsubtype='range'),
                f_token(tvalue=',', ttype='operator-infix', tsubtype='union'),
                f_token(tvalue='A2:B2', ttype='operand', tsubtype='range')
            ]
        )

    def test_range_address_function(self):
        self.assertASTNodesEqual(
            self.parse('SUM(A1:B1)'),
            [
                f_token(tvalue='SUM', ttype='function', tsubtype='start'),
                f_token(tvalue='A1:B1', ttype='operand', tsubtype='range'),
                f_token(tvalue='', ttype='function', tsubtype='stop'),
            ]
        )

    def test_array(self):
        self.assertASTNodesEqual(
            self.parse('{{1, 2}, {3, 4}}'),
            [
                f_token(tvalue='ARRAY', ttype='function', tsubtype='start'),
                f_token(tvalue='ARRAYROW', ttype='function', tsubtype='start'),
                f_token(tvalue='ARRAY', ttype='function', tsubtype='start'),
                f_token(tvalue='ARRAYROW', ttype='function', tsubtype='start'),
                f_token(tvalue='1', ttype='operand', tsubtype='number'),
                f_token(tvalue=',', ttype='argument', tsubtype=''),
                f_token(tvalue='2', ttype='operand', tsubtype='number'),
                f_token(tvalue='', ttype='function', tsubtype='stop'),
                f_token(tvalue='', ttype='function', tsubtype='stop'),
                f_token(tvalue=',', ttype='argument', tsubtype=''),
                f_token(tvalue='ARRAY', ttype='function', tsubtype='start'),
                f_token(tvalue='ARRAYROW', ttype='function', tsubtype='start'),
                f_token(tvalue='3', ttype='operand', tsubtype='number'),
                f_token(tvalue=',', ttype='argument', tsubtype=''),
                f_token(tvalue='4', ttype='operand', tsubtype='number'),
                f_token(tvalue='', ttype='function', tsubtype='stop'),
                f_token(tvalue='', ttype='function', tsubtype='stop'),
                f_token(tvalue='', ttype='function', tsubtype='stop'),
                f_token(tvalue='', ttype='function', tsubtype='stop'),
            ]
        )

    def test_int(self):
        self.assertASTNodesEqual(
            self.parse('1'),
            [
                f_token(tvalue='1', ttype='operand', tsubtype='number'),
            ]
        )

    def test_float(self):
        self.assertASTNodesEqual(
            self.parse('1.0'),
            [
                f_token(tvalue='1.0', ttype='operand', tsubtype='number'),
            ]
        )

    def test_text(self):
        self.assertASTNodesEqual(
            self.parse('"data"'),
            [
                f_token(tvalue='data', ttype='operand', tsubtype='text'),
            ]
        )

    def test_bool(self):
        self.assertASTNodesEqual(
            self.parse('TRUE'),
            [
                f_token(tvalue='TRUE', ttype='operand', tsubtype='logical'),
            ]
        )

    def test_null(self):
        self.assertASTNodesEqual(
            self.parse('NULL'),
            [
                f_token(tvalue='NULL', ttype='operand', tsubtype='range'),
            ]
        )

    def test_error(self):
        self.assertASTNodesEqual(
            self.parse('#NUM!'),
            [
                f_token(tvalue='#NUM!', ttype='operand', tsubtype='error'),
            ]
        )

    def test_concat(self):
        self.assertASTNodesEqual(
            self.parse('"hi" & "de"'),
            [
                f_token(tvalue='hi', ttype='operand', tsubtype='text'),
                f_token(
                    tvalue='&', ttype='operator-infix',
                    tsubtype='concatenate'),
                f_token(tvalue='de', ttype='operand', tsubtype='text'),
            ]
        )

    def test_array_formula(self):
        self.assertASTNodesEqual(
            self.parse('{=1}'),
            [
                f_token(tvalue='ARRAY', ttype='function', tsubtype='start'),
                f_token(tvalue='ARRAYROW', ttype='function', tsubtype='start'),
                f_token(
                    tvalue='=', ttype='operator-infix', tsubtype='logical'),
                f_token(tvalue='1', ttype='operand', tsubtype='number'),
                f_token(tvalue='', ttype='function', tsubtype='stop'),
                f_token(tvalue='', ttype='function', tsubtype='stop'),
            ]
        )



from .. import testing


class YearfracTest(testing.FunctionalTestCase):
    filename = "YEARFRAC.xlsx"

    def test_evaluation_A1(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A1')
        value = self.evaluator.evaluate('Sheet1!A1')
        self.assertEqualTruncated(excel_value, value)

    # only close within 3 decimal places.
    def test_evaluation_B1(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!B1')
        value = self.evaluator.evaluate('Sheet1!B1')
        self.assertEqualTruncated(excel_value, value, 3)

    def test_evaluation_C1(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!C1')
        value = self.evaluator.evaluate('Sheet1!C1')
        self.assertEqualTruncated(excel_value, value)



from .. import testing


class NPVTest(testing.FunctionalTestCase):
    filename = "XNPV.xlsx"

    def test_evaluation_A1(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A1')
        value = self.evaluator.evaluate('Sheet1!A1')
        self.assertEqualTruncated(excel_value, value)



import re

from . import operator, xlerrors, func_xltypes

CRITERIA_REGEX = r'(\W*)(.*)'

CRITERIA_OPERATORS = {
    '<': operator.OP_LT,
    '<=': operator.OP_LE,
    '=': operator.OP_EQ,
    '<>': operator.OP_NE,
    '>=': operator.OP_GE,
    '>': operator.OP_GT,
}


def parse_criteria(criteria):

    if isinstance(criteria, (str, func_xltypes.Text)):
        search = re.search(CRITERIA_REGEX, str(criteria)).group
        str_operator, str_value = search(1), search(2)

        operator = CRITERIA_OPERATORS.get(str_operator)
        if operator is None:
            operator = CRITERIA_OPERATORS['=']
            str_value = criteria

        value = str_value
        for XlType in (
            func_xltypes.Number,
            func_xltypes.DateTime,
            func_xltypes.Boolean
        ):
            try:
                value = XlType.cast(str_value)
            except xlerrors.ValueExcelError:
                pass
            else:
                break

        def check(probe):
            return operator(probe, value)

        return check

    criteria = func_xltypes.ExcelType.cast_from_native(criteria)

    # No support for arrays right now.
    if isinstance(criteria, func_xltypes.Array):
        raise xlerrors.ValueExcelError('Array criteria not supported.')

    def check(x):
        return x == criteria

    return check



from .. import testing

from xlcalculator.xlfunctions import xlerrors


class SqrtTest(testing.FunctionalTestCase):
    filename = "SQRT.xlsx"

    def test_evaluation_A1(self):
        self.assertEqual(
            self.evaluator.get_cell_value('Sheet1!A1'),
            self.evaluator.evaluate('Sheet1!A1')
        )

    def test_evaluation_B1(self):
        self.assertIsInstance(
            self.evaluator.evaluate('Sheet1!B1'), xlerrors.NumExcelError)

    def test_evaluation_C1(self):
        self.assertEqual(
            self.evaluator.get_cell_value('Sheet1!C1'),
            self.evaluator.evaluate('Sheet1!C1')
        )



from . import xl, xlerrors, func_xltypes


@xl.register()
@xl.validate_args
def CHOOSE(
        index_num: func_xltypes.XlNumber,
        *values,
) -> func_xltypes.XlAnything:
    """Uses index_num to return a value from the list of value arguments.

    https://support.office.com/en-us/article/
        choose-function-fc5c184f-cb62-4ec7-a46e-38653b98f5bc
    """
    if index_num <= 0 or index_num > 254:
        raise xlerrors.ValueExcelError(
            f"`index_num` {index_num} must be between 1 and 254")

    if index_num > len(values):
        raise xlerrors.ValueExcelError(
            f"`index_num` {index_num} must not be larger than the number of "
            f"values: {len(values)}")

    idx = int(index_num) - 1
    return values[idx]


@xl.register()
@xl.validate_args
def VLOOKUP(
        lookup_value: func_xltypes.XlAnything,
        table_array: func_xltypes.XlArray,
        col_index_num: func_xltypes.XlNumber,
        range_lookup=False
) -> func_xltypes.XlAnything:
    """Looks in the first column of an array and moves across the row to
    return the value of a cell.

    https://support.office.com/en-us/article/
        vlookup-function-0bbc8083-26fe-4963-8ab8-93a18ad188a1
    """
    if range_lookup:
        raise NotImplementedError("Excact match only supported at the moment.")

    col_index_num = int(col_index_num)

    if col_index_num > len(table_array.values[0]):
        raise xlerrors.ValueExcelError(
            'col_index_num is greater than the number of cols in table_array')

    table_array = table_array.set_index(0)

    if lookup_value not in table_array.index:
        raise xlerrors.NaExcelError(
            '`lookup_value` not in first column of `table_array`.')

    return table_array.loc[lookup_value].values[0]


@xl.register()
@xl.validate_args
def MATCH(
        lookup_value: func_xltypes.XlAnything,
        lookup_array: func_xltypes.XlArray,
        match_type: func_xltypes.XlAnything = 1,
) -> func_xltypes.XlAnything:
    assert len(lookup_array.values[0]) == 1

    lookup_array = lookup_array.flat

    if match_type == 1:
        if lookup_array != sorted(lookup_array):
            return xlerrors.NaExcelError(
                "Values must be sorted in ascending order"
            )
    if match_type == -1:
        if lookup_array != sorted(lookup_array, reverse=True):
            return xlerrors.NaExcelError(
                "Values must be sorted in descending order"
            )

    for i, val in enumerate(lookup_array):
        if val == lookup_value:
            return i + 1
        if match_type == 1 and val > lookup_value:
            return i or xlerrors.NaExcelError(
                "No lesser value found."
            )
        if match_type == -1 and val < lookup_value:
            return i or xlerrors.NaExcelError(
                "No greater value found."
            )
    return xlerrors.NaExcelError("No match found.")



import jsonpickle
import unittest

from xlcalculator import reader, xltypes, tokenizer
from . import testing


class ReaderTest(unittest.TestCase):

    def setUp(self):
        infile = open(testing.get_resource("reader.json"), "rb")
        json_bytes = infile.read()
        infile.close()
        data = jsonpickle.decode(
            json_bytes, keys=True,
            classes=(
                xltypes.XLCell, xltypes.XLFormula, xltypes.XLRange,
                tokenizer.f_token
            )
        )
        self.cells = data['cells']
        self.defined_names = data['defined_names']
        self.ranges = data['ranges']
        self.formulae = data['formulae']

    def test_read_cells(self):
        archive = reader.Reader(testing.get_resource("reader.xlsm"))
        archive.read()
        cells, formulae, ranges = \
            archive.read_cells(ignore_sheets=['Eleventh'])

        self.assertEqual(sorted(self.cells.keys()), sorted(cells.keys()))

    @unittest.skip(
        "Need to re-create reader.json as some elements of dynamic arrays "
        "are not being read correctly")
    def test_read_formulae(self):
        archive = reader.Reader(testing.get_resource("reader.xlsm"))
        archive.read()
        cells, formulae, ranges = \
            archive.read_cells(ignore_sheets=['Eleventh'])

        self.assertEqual(sorted(self.formulae.keys()), sorted(formulae.keys()))

    def test_read_defined_names(self):
        archive = reader.Reader(testing.get_resource("reader.xlsm"))
        archive.read()
        defined_names = archive.read_defined_names()

        its_a_blank = (
            set(sorted(defined_names.keys()))
            - set(sorted(self.defined_names.keys()))
        )
        self.assertEqual(list(its_a_blank), ['Its_a_blank'])



import functools
import inspect
import typing

from . import func_xltypes, xlerrors

COMPATIBILITY = 'EXCEL'
CELL_CHARACTER_LIMIT = 32767

TYPE_TO_CAST = {
    func_xltypes.XlNumber: func_xltypes.Number.cast,
    func_xltypes.XlText: func_xltypes.Text.cast,
    func_xltypes.XlBoolean: func_xltypes.Boolean.cast,
    func_xltypes.XlDateTime: func_xltypes.DateTime.cast,
    func_xltypes.XlArray: func_xltypes.Array.cast,
    func_xltypes.XlExpr: func_xltypes.Expr.cast,
    func_xltypes.XlAnything: func_xltypes.ExcelType.cast_from_native,
}


class Functions(dict):

    def register(self, func, name=None):
        if name is None:
            name = func.__name__
        self[name] = func

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)


FUNCTIONS = Functions()


def register(name=None):
    """Decorator to register a function."""

    def registerFunction(func):
        FUNCTIONS.register(func, name)
        return func

    return registerFunction


def _validate(vtype, val, name):
    cast = TYPE_TO_CAST.get(vtype, None)
    if cast is not None:
        return cast(val)

    # Support lists with value types
    if getattr(vtype, '__origin__', None) in [list, tuple]:
        itype = vtype.__args__[0]
        if itype != func_xltypes.XlArray:
            val = flatten(val)
        return tuple(filter(
            lambda x: x is not None,
            [_safe_validate(itype, item, name) for item in val]
        ))

    # Support unions
    if getattr(vtype, '__origin__', None) == typing.Union:
        for stype in vtype.__args__:
            try:
                return _validate(stype, val, name)
            except xlerrors.ExcelError:
                pass
        raise xlerrors.ValueExcelError(val)

    return val


def _safe_validate(vtype, val, name):
    try:
        return _validate(vtype, val, name)
    except xlerrors.ExcelError:
        return None


def validate_args(func):

    @functools.wraps(func)
    def validate(*args, **kw):
        sig = inspect.signature(func)
        bound = sig.bind(*args, **kw)
        # 1. Convert all input parameters to Excel Types.
        for pname, value in list(bound.arguments.items()):
            if isinstance(value, xlerrors.ExcelError):
                return value
            try:
                bound.arguments[pname] = _validate(
                    sig.parameters[pname].annotation, value, pname)
            except xlerrors.ExcelError as err:
                return err
        # 2. Run the function to compute the result.
        try:
            res = func(*bound.args, **bound.kwargs)
        except xlerrors.ExcelError as err:
            # Never crash on Excel errors as we want to store them as the cell
            # value.
            return err
        # 3. Convert the result to an Excel type.
        return _validate(sig.return_annotation, res, 'return')

    return validate


def flatten(values):
    """Fully recursive flattening."""
    flat = []
    if isinstance(values, func_xltypes.Array):
        values = values.flat
    for value in values:
        if isinstance(value, func_xltypes.Array):
            flat.extend(value.flat)
        elif isinstance(value, (list, tuple)):
            flat.extend(flatten(value))
        else:
            flat.append(value)
    return flat


def length(values):
    return len(flatten(values))



from typing import Tuple

from . import xl, xlerrors, func_xltypes


@xl.register()
@xl.validate_args
def CONCAT(
        *texts: Tuple[func_xltypes.XlText]
) -> func_xltypes.XlText:
    """The CONCAT function combines the text from multiple ranges and/or
    strings, but it doesn't provide delimiter or IgnoreEmpty arguments.

    https://support.office.com/en-us/article/
        concat-function-9b1a9a3f-94ff-41af-9736-694cbd6b4ca2
    """
    if len(texts) > 254:
        raise xlerrors.ValueExcelError(
            f"Can't concat more than 254 arguments. Provided: {len(texts)}")

    texts = xl.flatten(texts)
    return ''.join([
        str(text) for text in xl.flatten(texts)
    ])


@xl.register()
@xl.validate_args
def CONCATENATE(
        *parameters: Tuple[func_xltypes.XlAnything]
) -> func_xltypes.XlText:
    """Use CONCATENATE, one of the text functions, to join two or more
    text strings into one string.

    https://support.microsoft.com/en-us/office/
        concatenate-function-8f8ae884-2ca8-4f7a-b093-75d702bea31d

    IMPORTANT: In Excel 2016, Excel Mobile, and Excel for the web, this
    function has been replaced with the CONCAT function. Although the
    CONCATENATE function is still available for backward compatibility,
    you should consider using CONCAT from now on. This is because CONCATENATE
    may not be available in future versions of Excel.
    """

    return CONCAT(
        [func_xltypes.Text.cast(parameter) for parameter in parameters]
    )


@xl.register()
@xl.validate_args
def EXACT(
        text1: func_xltypes.XlText,
        text2: func_xltypes.XlText
) -> func_xltypes.XlBoolean:
    """Compares two text strings and returns TRUE if they are exactly the
    same, FALSE otherwise. EXACT is case-sensitive but ignores formatting
    differences.

    https://support.microsoft.com/en-us/office/
        exact-function-d3087698-fc15-4a15-9631-12575cf29926
    """

    return str(text1) == str(text2)


@xl.register()
@xl.validate_args
def FIND(
        find_text: func_xltypes.XlText,
        within_text: func_xltypes.XlText,
        start_num: func_xltypes.Number = 0,
) -> func_xltypes.Number:
    """FIND and FINDB locate one text string within a second text string,
    and return the number of the starting position of the first text string
    from the first character of the second text string.

    https://support.microsoft.com/en-us/office/
        find-findb-functions-c7912941-af2a-4bdf-a553-d0d89b0a0628
    """
    index = 1
    within_text_str = str(within_text)
    find_text_str = str(find_text)
    start_num_int = int(start_num)

    # Excel operates in 1-based land, Python is usually 0-based.
    if start_num_int > 0:
        start_num_int = start_num_int - 1

    try:
        index = within_text_str.index(find_text_str, start_num_int) + 1
    except ValueError:
        raise xlerrors.ValueExcelError(
            f"Text {find_text} isn't found in"
            f" {within_text_str[:start_num_int]}")

    return index


@xl.register()
@xl.validate_args
def LEFT(
        text: func_xltypes.XlText,
        num_chars: func_xltypes.XlNumber = 1
) -> func_xltypes.XlText:
    """LEFT returns the first character or characters in a text string,
    based on the number of characters you specify.

    https://support.office.com/en-us/article/
        left-leftb-functions-9203d2d2-7960-479b-84c6-1ea52b99640c
    """
    return str(text)[:int(num_chars)]


@xl.register()
@xl.validate_args
def LEN(
        text: func_xltypes.XlText
) -> func_xltypes.XlNumber:
    """LEN returns the number of characters in a text string.

    https://support.office.com/en-us/article/
        len-lenb-functions-29236f94-cedc-429d-affd-b5e33d2c67cb
    """
    return len(str(text))


@xl.register()
@xl.validate_args
def LOWER(
        text: func_xltypes.XlText
) -> func_xltypes.XlText:
    """Converts all uppercase letters in a text string to lowercase.

    https://support.office.com/en-us/article/
        lower-function-3f21df02-a80c-44b2-afaf-81358f9fdeb4
    """
    return str(text).lower()


@xl.register()
@xl.validate_args
def MID(
        text: func_xltypes.XlText,
        start_num: func_xltypes.Number,
        num_chars: func_xltypes.Number
) -> func_xltypes.XlText:
    """Returns a specific number of characters from a text string, starting
    at the position you specify, based on the number of characters you specify.

    https://support.office.com/en-us/article/
        mid-midb-functions-d5f9e25c-d7d6-472e-b568-4ecb12433028
    """
    text = str(text)

    if len(text) > xl.CELL_CHARACTER_LIMIT:
        raise xlerrors.ValueExcelError(
            f'Text is too long. Is {len(text)} but needs to '
            f'be {xl.CELL_CHARACTER_LIMIT} or less.')

    start_num = int(start_num)
    if start_num < 1:
        raise xlerrors.NumExcelError(f'{start_num} is < 1')

    num_chars = int(num_chars)
    if num_chars < 0:
        raise xlerrors.NumExcelError(f'{num_chars} is < 0')

    start_idx = start_num - 1
    return text[start_idx:start_idx + num_chars]


@xl.register()
@xl.validate_args
def REPLACE(
        old_text: func_xltypes.XlText,
        start_num: func_xltypes.XlNumber,
        num_chars: func_xltypes.XlNumber,
        new_text: func_xltypes.XlText
) -> func_xltypes.XlText:
    """REPLACE replaces part of a text string, based on the number of
    characters you specify, with a different text string.

    https://support.office.com/en-us/article/
        replace-replaceb-functions-8d799074-2425-4a8a-84bc-82472868878a
    """
    old_text_str = str(old_text)
    start_num_int = int(start_num) - 1  # Excel is 1-based, Python is 0-based
    num_chars_int = int(num_chars)
    new_text_str = str(new_text)

    sliced_old_text = old_text_str[start_num_int:
                                   start_num_int + num_chars_int]

    return old_text_str.replace(sliced_old_text, new_text_str)


@xl.register()
@xl.validate_args
def RIGHT(
        text: func_xltypes.XlText,
        num_chars: func_xltypes.XlNumber = 1
) -> func_xltypes.XlText:
    """Returns the last character or characters in a text string.

    https://support.office.com/en-us/article/
        right-rightb-functions-240267ee-9afa-4639-a02b-f19e1786cf2f
    """
    return str(text)[-int(num_chars):]


@xl.register()
@xl.validate_args
def TRIM(
        text: func_xltypes.XlText
) -> func_xltypes.XlText:
    """Removes all spaces from text except for single spaces between words.

    https://support.office.com/en-us/article/
        trim-function-410388fa-c5df-49c6-b16c-9e5630b479f9
    """
    return str(text).strip()


@xl.register()
@xl.validate_args
def UPPER(
        text: func_xltypes.XlText
) -> func_xltypes.XlText:
    """Converts text to uppercase.

    https://support.office.com/en-us/article/
        upper-function-c11f29b3-d1a3-4537-8df6-04d0049963d6
    """
    return str(text).upper()



from . import ast_nodes, tokenizer


class Operator(object):
    """Small wrapper class to manage operators during shunting yard"""

    def __init__(self, value, precedence, associativity):
        self.value = value
        self.precedence = precedence
        self.associativity = associativity


# http://office.microsoft.com/en-us/excel-help/
#    calculation-operators-and-precedence-HP010078886.aspx
OPERATORS = {
    ':': Operator(':', 8, 'left'),
    '': Operator(' ', 8, 'left'),
    ',': Operator(',', 8, 'left'),
    'u-': Operator('u-', 7, 'right'),  # unary negation
    '%': Operator('%', 6, 'left'),
    '^': Operator('^', 5, 'left'),
    '*': Operator('*', 4, 'left'),
    '/': Operator('/', 4, 'left'),
    '+': Operator('+', 3, 'left'),
    '-': Operator('-', 3, 'left'),
    '&': Operator('&', 2, 'left'),
    '=': Operator('=', 1, 'left'),
    '<': Operator('<', 1, 'left'),
    '>': Operator('>', 1, 'left'),
    '<=': Operator('<=', 1, 'left'),
    '>=': Operator('>=', 1, 'left'),
    '<>': Operator('<>', 1, 'left'),
}


class FormulaParser:
    """Excel Formula Parser"""

    def parse(self, formula, named_ranges=None, tokenize_range=False):
        """Parse formula into evaluable AST.
        """
        # 1. Parse the formula into syntactic tokens.
        tokens = self.tokenize(formula)
        # 2. Organize tokens into reverse polish notation.
        nodes = self.shunting_yard(tokens, named_ranges)
        # 3. Construct the AST by connecting the AST nodes.
        ast = self.build_ast(nodes)
        return ast

    def tokenize(self, formula, tokenize_range=False):
        # Remove leading "=" sign.
        if formula.startswith('='):
            formula = formula[1:]

        excel_parser = tokenizer.ExcelParser(tokenize_range=tokenize_range)
        return excel_parser.parse(formula).items

    def shunting_yard(self, raw_tokens, named_ranges, tokenize_range=False):
        """Reorganize tokens into proper reverse polish notation.

        Core algorithm taken from wikipedia with varargs extensions from

        http://www.kallisti.net.nz/blog/2008/02/
            extension-to-the-shunting-yard-algorithm-to-allow-
            variable-numbers-of-arguments-to-functions/


        The ref is the cell address which is passed down to the actual
        compiled python code.  Range basic operations signature require this
        reference, so it has to be written during OperatorNode.emit()

          https://github.com/iOiurson/koala/blob/master/koala/ast/graph.py#L292.

        This is needed because Excel range basic operations (+, -, * ...) are
        applied on matching cells.

        Example:

        Cell C2 has the following formula 'A1:A3 + B1:B3'.  The output will
        actually be A2 + B2, because the formula is relative to cell C2.
        """
        # insert tokens for '(' and ')', to make things clearer below
        tokens = []
        for token in raw_tokens:
            if token.ttype == "function" and token.tsubtype == "start":
                token.tsubtype = ""
                tokens.append(token)
                tokens.append(tokenizer.f_token('(', 'arglist', 'start'))

            elif token.ttype == "function" and token.tsubtype == "stop":
                tokens.append(tokenizer.f_token(')', 'arglist', 'stop'))

            elif token.ttype == "subexpression" and token.tsubtype == "start":
                token.tvalue = '('
                tokens.append(token)

            elif token.ttype == "subexpression" and token.tsubtype == "stop":
                token.tvalue = ')'
                tokens.append(token)

            elif (token.ttype == "operand" and token.tsubtype == "range"
                    and token.tvalue in named_ranges):
                # Resolve the named range once and for all.
                token.tvalue = named_ranges[token.tvalue]
                tokens.append(token)

            else:
                tokens.append(token)

        output = []
        stack = []
        were_values = []
        arg_count = []

        new_tokens = []

        # reconstruct expressions with ':' and replace the corresponding
        # tokens by the reconstructed expression
        if not tokenize_range:
            for index, token in enumerate(tokens):
                new_tokens.append(token)

                if type(token.tvalue) == str:

                    # example -> :OFFSET( or simply :A10
                    if token.tvalue.startswith(':'):
                        depth = 0
                        expr = ''
                        rev = reversed(tokens[:index])

                        # going backwards, 'stop' starts, 'start' stops
                        for reversed_token in rev:
                            if reversed_token.tsubtype == 'stop':
                                depth += 1

                            elif (
                                depth > 0
                                and reversed_token.tsubtype == 'start'
                            ):
                                depth -= 1

                            expr = reversed_token.tvalue + expr

                            new_tokens.pop()

                            if depth == 0:
                                # these 2 lines are needed to remove INDEX()
                                new_tokens.pop()
                                new_tokens.pop()
                                expr = rev.next().tvalue + expr
                                break

                        expr += token.tvalue

                        depth = 0

                        if token.tvalue[1:] in ['OFFSET', 'INDEX']:
                            for t in tokens[(index + 1):]:
                                if t.tsubtype == 'start':
                                    depth += 1

                                elif depth > 0 and t.tsubtype == 'stop':
                                    depth -= 1

                                expr += t.tvalue
                                tokens.remove(t)

                                if depth == 0:
                                    break

                        new_tokens.append(
                            tokenizer.f_token(expr, 'operand', 'pointer'))

                    # example -> A1:OFFSET(
                    elif ':OFFSET' in token.tvalue or ':INDEX' in token.tvalue:
                        depth = 0
                        expr = ''
                        expr += token.tvalue

                        for t in tokens[(index + 1):]:
                            if t.tsubtype == 'start':
                                depth += 1

                            elif t.tsubtype == 'stop':
                                depth -= 1

                            expr += t.tvalue
                            tokens.remove(t)

                            if depth == 0:
                                new_tokens.pop()
                                break

                        new_tokens.append(
                            tokenizer.f_token(expr, 'operand', 'pointer'))

        tokens = new_tokens if new_tokens else tokens

        for token in tokens:
            if token.ttype == "operand":
                output.append(self.create_node(token))

                if were_values:
                    were_values.pop()
                    were_values.append(True)

            elif token.ttype == "function":
                stack.append(token)
                arg_count.append(0)

                if were_values:
                    were_values.pop()
                    were_values.append(True)

                were_values.append(False)

            elif token.ttype == "argument":

                while stack and (stack[-1].tsubtype != "start"):
                    output.append(self.create_node(stack.pop()))

                if were_values.pop():
                    arg_count[-1] += 1
                were_values.append(False)

                if not len(stack):
                    raise ValueError("Mismatched or misplaced parentheses")

            elif token.ttype.startswith('operator'):

                if token.ttype.endswith('-prefix') and token.tvalue == "-":
                    o1 = OPERATORS['u-']

                else:
                    o1 = OPERATORS[token.tvalue]

                while stack and stack[-1].ttype.startswith('operator'):
                    if (
                            stack[-1].ttype.endswith('-prefix')
                            and stack[-1].tvalue == "-"
                    ):
                        o2 = OPERATORS['u-']

                    else:
                        o2 = OPERATORS[stack[-1].tvalue]

                    if (
                            (o1.associativity == "left"
                             and o1.precedence <= o2.precedence)
                            or (o1.associativity == "right"
                                and o1.precedence < o2.precedence)
                    ):
                        output.append(
                            self.create_node(stack.pop()))

                    else:
                        break

                stack.append(token)

            elif token.tsubtype == "start":
                stack.append(token)

            elif token.tsubtype == "stop":
                while stack and stack[-1].tsubtype != "start":
                    output.append(self.create_node(stack.pop()))

                if not stack:
                    raise SyntaxError("Mismatched or misplaced parentheses")

                stack.pop()

                if stack and stack[-1].ttype == "function":
                    f = self.create_node(stack.pop())
                    a = arg_count.pop()
                    w = were_values.pop()
                    if w:
                        a += 1
                    f.num_args = a
                    output.append(f)

        while stack:
            if (stack[-1].tsubtype == "start" or stack[-1].tsubtype == "stop"):
                raise SyntaxError("Mismatched or misplaced parentheses")

            output.append(self.create_node(stack.pop()))

        # convert to list
        return [x for x in output]

    def create_node(self, token):
        if token.ttype == "operand":
            if token.tsubtype in ["range", "pointer"]:
                return ast_nodes.RangeNode(token)
            else:
                return ast_nodes.OperandNode(token)

        elif token.ttype == "function":
            return ast_nodes.FunctionNode(token)

        elif token.ttype.startswith("operator"):
            return ast_nodes.OperatorNode(token)

        else:
            raise ValueError('Unknown token type: ' + token.ttype)

    def build_ast(self, nodes):
        """Update AST nodes to build a proper parse tree.

        XXX: There is really no need for this. The shunting yeard algorithm
        should jsut take care of it.
        """
        stack = []

        for node in nodes:
            if isinstance(node, ast_nodes.OperatorNode):
                if node.ttype == "operator-infix":
                    # Stack has arguments in reverse order.
                    node.right = stack.pop()
                    node.left = stack.pop()
                else:
                    node.right = stack.pop()

            elif isinstance(node, ast_nodes.FunctionNode):
                args = []
                for _ in range(node.num_args):
                    args.append(stack.pop())
                # Stack has arguments in reverse order.
                node.args = list(reversed(args))

            stack.append(node)

        return stack.pop()



import contextlib
import mock
import openpyxl


class Cell(openpyxl.cell.cell.Cell):

    __slots__ = openpyxl.cell.cell.Cell.__slots__ + (
        'cvalue',
    )


class WorkSheetParser(openpyxl.worksheet._reader.WorkSheetParser):

    def parse_cell(self, element):
        cell = super().parse_cell(element)
        if cell['data_type'] == 'f':
            with mock.patch.object(self, 'data_only', True):
                cell_data = super().parse_cell(element)
            cell['cvalue'] = cell_data.get('value')
        return cell


class WorksheetReader(openpyxl.worksheet._reader.WorksheetReader):

    def __init__(self, ws, xml_source, shared_strings, data_only):
        super().__init__(ws, xml_source, shared_strings, data_only)
        self.parser = WorkSheetParser(
            xml_source, shared_strings, data_only, ws.parent.epoch,
            ws.parent._date_formats)

    def bind_cells(self):
        for idx, row in self.parser.parse():
            for cell in row:
                style = self.ws.parent._cell_styles[cell['style_id']]
                c = Cell(
                    self.ws, row=cell['row'], column=cell['column'],
                    style_array=style)
                c._value = cell['value']
                c.data_type = cell['data_type']
                # xlcalculator extension: Store the cached value as well.
                if c.data_type == 'f':
                    c.cvalue = cell['cvalue']
                self.ws._cells[(cell['row'], cell['column'])] = c
        self.ws.formula_attributes = self.parser.array_formulae
        if self.ws._cells:
            self.ws._current_row = self.ws.max_row


@contextlib.contextmanager
def openpyxl_WorksheetReader_patch():
    """Allow extraction of cached formula value."""
    with mock.patch.object(
            openpyxl.worksheet._reader, 'WorksheetReader', WorksheetReader):
        with mock.patch.object(
                openpyxl.reader.excel, 'WorksheetReader', WorksheetReader):
            yield



from .. import testing


class CEILINGTest(testing.FunctionalTestCase):

    filename = "CEILING.xlsx"

    def test_evaluation_A2(self):
        self.assertEqual(
            self.evaluator.evaluate('Sheet1!A2'),
            self.evaluator.get_cell_value('Sheet1!A2')
        )

    def test_evaluation_A3(self):
        self.assertEqual(
            self.evaluator.evaluate('Sheet1!A3'),
            self.evaluator.get_cell_value('Sheet1!A3')
        )

    def test_evaluation_A4(self):
        self.assertEqual(
            self.evaluator.evaluate('Sheet1!A4'),
            self.evaluator.get_cell_value('Sheet1!A4')
        )

    def test_evaluation_A5(self):
        self.assertEqual(
            self.evaluator.evaluate('Sheet1!A5'),
            self.evaluator.get_cell_value('Sheet1!A5')
        )

    def test_evaluation_A6(self):
        self.assertEqual(
            self.evaluator.evaluate('Sheet1!A6'),
            self.evaluator.get_cell_value('Sheet1!A6')
        )

    def test_evaluation_A7(self):
        self.assertEqual(
            self.evaluator.evaluate('Sheet1!A7'),
            self.evaluator.get_cell_value('Sheet1!A7')
        )

    def test_evaluation_A8(self):
        self.assertEqual(
            self.evaluator.evaluate('Sheet1!A8'),
            self.evaluator.get_cell_value('Sheet1!A8')
        )

    def test_evaluation_A9(self):
        self.assertEqual(
            self.evaluator.evaluate('Sheet1!A9'),
            self.evaluator.get_cell_value('Sheet1!A9')
        )



import datetime
import dateutil
import numpy
import pandas
from typing import Optional, Union, NewType

from . import utils, xlerrors

NATIVE_TO_XLTYPE = {}


def register(cls):
    for native_type in cls.native_types:
        NATIVE_TO_XLTYPE[native_type] = cls
    return cls


class ExcelType:

    __slots__ = ('value')

    native_types = ()

    sort_precedence = 0

    def __new__(cls, value):
        inst = super().__new__(cls)
        assert isinstance(value, cls.native_types), value
        inst.value = value
        return inst

    @classmethod
    def cast(cls, value):
        if isinstance(value, cls):
            return value
        if not isinstance(value, ExcelType):
            if type(value) not in NATIVE_TO_XLTYPE:
                raise xlerrors.ValueExcelError(
                    f'Unknown object type: {type(value)} ({value})')
            value = NATIVE_TO_XLTYPE[type(value)](value)
        return getattr(value, f'__{cls.__name__}__')()

    @classmethod
    def is_type(cls, value):
        return isinstance(value, (cls,) + cls.native_types)

    @classmethod
    def cast_from_native(cls, value):
        if isinstance(value, xlerrors.ExcelError):
            return value
        if isinstance(value, tuple(NATIVE_TO_XLTYPE.values())):
            return value
        return NATIVE_TO_XLTYPE[type(value)](value)

    def _sort_key(self, other):
        return (self.sort_precedence, self.value)

    def __add__(self, other):
        return Number(Number.cast(self).value + Number.cast(other).value)

    def __sub__(self, other):
        return Number(Number.cast(self).value - Number.cast(other).value)

    def __mul__(self, other):
        return Number(Number.cast(self).value * Number.cast(other).value)

    def __truediv__(self, other):
        ovalue = float(Number.cast(other))
        if ovalue == 0:
            raise xlerrors.DivZeroExcelError()
        return Number(float(Number.cast(self)) / ovalue)

    def __pow__(self, other):
        return Number(Number.cast(self).value ** Number.cast(other).value)

    def __and__(self, other):
        # Highjacking bitwise "and" to implement logical "and"
        return Boolean(bool(self) and bool(other))

    def __or__(self, other):
        # Highjacking bitwise "or" to implement logical "or"
        return Boolean(bool(self) or bool(other))

    __radd__ = __add__
    __rsub__ = __sub__
    __rmul__ = __mul__
    __rtruediv__ = __truediv__
    __rpow__ = __pow__
    __rand__ = __and__
    __ror__ = __or__

    def __lt__(self, other):
        other = ExcelType.cast_from_native(other)
        return Boolean(self._sort_key(other) < other._sort_key(self))

    def __le__(self, other):
        other = ExcelType.cast_from_native(other)
        return Boolean(self._sort_key(other) <= other._sort_key(self))

    def __eq__(self, other):
        other = ExcelType.cast_from_native(other)
        return Boolean(self._sort_key(other) == other._sort_key(self))

    def __ne__(self, other):
        other = ExcelType.cast_from_native(other)
        return Boolean(self._sort_key(other) != other._sort_key(self))

    def __gt__(self, other):
        other = ExcelType.cast_from_native(other)
        return Boolean(self._sort_key(other) > other._sort_key(self))

    def __ge__(self, other):
        other = ExcelType.cast_from_native(other)
        return Boolean(self._sort_key(other) >= other._sort_key(self))

    def __int__(self):
        try:
            return int(float(self.value))
        except ValueError:
            raise xlerrors.ValueExcelError(
                f'Could not convert {repr(self.value)} to int.')

    def __float__(self):
        try:
            return float(self.value)
        except (TypeError, ValueError):
            raise xlerrors.ValueExcelError(
                f'Could not convert {repr(self.value)} to float.')

    def __str__(self):
        return str(self.value)

    def __bool__(self):
        return bool(self.value)

    def __number__(self):
        return float(self.value)

    def __datetime__(self):
        raise NotImplementedError

    def __Number__(self):
        return Number(self.__number__())

    def __Text__(self):
        return Text(self.__str__())

    def __Boolean__(self):
        return Boolean(self.__bool__())

    def __DateTime__(self):
        return DateTime(self.__datetime__())

    def __Blank__(self):
        return self.__class__(None)

    def __hash__(self):
        return hash(self.value)

    def __repr__(self):
        return f'<{self.__class__.__name__} {repr(self.value)}>'


@register
class Number(ExcelType):

    native_types = (int, float, numpy.int64, numpy.float64)

    blank_value = 0

    @property
    def is_whole(self):
        return isinstance(self.value, int)

    @property
    def is_decimal(self):
        return isinstance(self.value, float)

    def __mod__(self, other):
        return Number(self.value % Number.cast(other).value)

    __rmod__ = __mod__

    def __neg__(self):
        return Number(self.value.__neg__())

    def __pos__(self):
        return Number(self.value.__pos__())

    def __invert__(self):
        return Number(-self.value)

    def __abs__(self):
        return Number(self.value.__abs__())

    def __round__(self, ndigits=None):
        return Number(self.value.__round__(ndigits))

    def __trunc__(self):
        return Number(self.value.__trunc__())

    def __number__(self):
        return self.value

    def __datetime__(self):
        return utils.number_to_datetime(self.value)

    def __Blank__(self):
        return self.__class__(0)


@register
class Text(ExcelType):

    boolean_texts = ['false', 'true']

    native_types = (str,)
    sort_precedence = 1

    def __number__(self):
        try:
            return int(self.value)
        except ValueError:
            pass
        try:
            return float(self.value)
        except ValueError:
            pass
        # For arithmetic, boolean text is actually interpreted.
        try:
            return int(self.__bool__(by_content_only=True))
        except xlerrors.ValueExcelError:
            pass
        # Try casting to datetime first, since the string might represent one
        # and that can be concerted.
        try:
            return utils.datetime_to_number(self.__datetime__())
        except xlerrors.ValueExcelError:
            pass
        raise xlerrors.ValueExcelError(
            f'Could not convert {repr(self.value)} to float.')

    def __bool__(self, by_content_only=False):
        if self.value.lower() in self.boolean_texts:
            return (self.value.lower() == 'true')
        if by_content_only:
            raise xlerrors.ValueExcelError(
                f'Could not convert {repr(self.value)} to bool.')
        return bool(self.value)

    def __Boolean__(self):
        return Boolean(self.__bool__(by_content_only=True))

    def __datetime__(self):
        try:
            return utils.number_to_datetime(float(self.value))
        except (ValueError, OverflowError):
            pass
        try:
            return dateutil.parser.parse(self.value)
        except (ValueError, OverflowError):
            pass
        raise xlerrors.ValueExcelError(
            f'Could not cast {repr(self.value)} (of type {type(self.value)} '
            f'to date/time.')

    def __Blank__(self):
        return self.__class__('')

    def __lt__(self, other):
        # Text is always greater for comparison and is not converted.
        if isinstance(other, DateTime):
            return Boolean(False)

        return Boolean(self.value.upper() < str(other).upper())

    def __le__(self, other):
        return Boolean(self.value.upper() <= str(other).upper())

    def __eq__(self, other):
        if self.value in (None, '') \
                and (other in (None, '') or isinstance(other, None)):
            return Boolean(True)

        return Boolean(self.value.upper() == str(other).upper())

    def __ne__(self, other):
        return Boolean(self.value.upper() != str(other).upper())

    def __gt__(self, other):
        return Boolean(self.value.upper() > str(other).upper())

    def __ge__(self, other):
        return Boolean(self.value.upper() >= str(other).upper())

    def __hash__(self):
        return hash(self.value)


@register
class Boolean(ExcelType):

    native_types = (bool,)
    sort_precedence = 2
    datetime_true = datetime.datetime(1999, 12, 31)
    datetime_false = datetime.datetime(1999, 12, 30)

    def _sort_key(self, other):
        return (self.sort_precedence, int(self.value))

    def __number__(self):
        return int(self.value)

    def __Boolean__(self):
        return self

    def __datetime__(self):
        return self.datetime_true if self.value else self.datetime_false

    def __Blank__(self):
        return self.__class__(False)


@register
class DateTime(ExcelType):

    native_types = (datetime.datetime, numpy.datetime64)

    def _sort_key(self, other):
        return self.__Number__()._sort_key(other)

    def __int__(self):
        return int(float(self))

    def __float__(self):
        return utils.datetime_to_number(self.value)

    __number__ = __float__

    def __DateTime__(self):
        return self

    def __Blank__(self):
        return None

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.value.isoformat()}>'


@register
class Blank(ExcelType):

    native_types = (type(None),)

    def __new__(cls, value=None):
        return super().__new__(cls, None)

    @classmethod
    def is_blank(cls, value):
        return isinstance(value, (cls,) + cls.native_types) or value == ''

    def _sort_key(self, other):
        return other.__Blank__()._sort_key(self)

    def __and__(self, other):
        if isinstance(other, self.native_types + (Blank,)):
            raise xlerrors.ValueExcelError('Cannot AND two blank values.')
        return Boolean.cast(other)

    def __or__(self, other):
        if isinstance(other, self.native_types + (Blank,)):
            raise xlerrors.ValueExcelError('Cannot OR two blank values.')
        return Boolean.cast(other)

    def __number__(self):
        return 0.0

    def __int__(self):
        return 0

    def __float__(self):
        return 0.0

    def __bool__(self):
        return False

    def __str__(self):
        return ''

    def __repr__(self):
        return '<BLANK>'


BLANK = Blank()


def _safe_cast(func, empty_value=None):

    def safe_cast(value):
        try:
            return func(value)
        except xlerrors.ExcelError:
            return empty_value

    return safe_cast


def _convert_nested_list(value):
    if not isinstance(value, (list, tuple)):
        try:
            return ExcelType.cast_from_native(value)
        except Exception:
            # There are cases where internal pandas data structures are passed
            # around.
            return value
    return [_convert_nested_list(item) for item in value]


@register
class Array(pandas.DataFrame):

    native_types = (list, tuple)

    def __init__(self, data, *args, **kw):
        try:
            super().__init__(_convert_nested_list(data), *args, **kw)
        except ValueError:
            raise xlerrors.ValueExcelError(f'Invalid array argument: {data}')

    @property
    def _constructor(self):
        return Array

    @property
    def flat(self):
        return list(self.values.flat)

    @classmethod
    def cast(cls, value):
        if isinstance(value, cls):
            return value
        if not isinstance(value, cls.native_types):
            value = [[value]]
        return Array(value)

    def flatten(self, xltype=None, filt=None):
        cast = _safe_cast(Number.cast, None) \
            if xltype is not None else lambda x: x
        return list(filter(filt, [cast(item) for item in self.values.flat]))

    def cast_to_numbers(self):
        return self.applymap(_safe_cast(Number.cast, Number(0.0)))

    def cast_to_booleans(self):
        return self.applymap(_safe_cast(Boolean.cast, Boolean(True)))

    def cast_to_texts(self):
        return self.applymap(_safe_cast(Text.cast, Text('')))


class Expr:
    """Expression

    Represents an expression that has yet to be evaluated. This class is
    agnostic to the internals of the implementation details. In other words,
    it does not expect particular objects. all it needs is a callable and its
    arguments.

    The constructor also accepts any arbitrary keywords that will be set on
    the expression to allow for debugging and discovery. Such info can include
    AST nodes, return type, cell reference, etc. (None of th einfo fields are
    required or will be considered by this library.)
    """

    def __init__(self, callable, args=(), kwargs={}, **info):
        self.callable = callable
        self.args = args
        self.kwargs = kwargs.copy()
        for name, value in info.items():
            setattr(self, name, value)

    @classmethod
    def cast(cls, value):
        if isinstance(value, cls):
            return value
        return ValueExpr(value)

    def __call__(self):
        return self.callable(*self.args, *self.kwargs)


def ValueExpr(value):
    return Expr(lambda: value, value=value)


# Python compatible types

# We can really anything into Excel arguments, so let's allow all those
# types.
_Anything = Optional[Union[
    int, float, bool, str, datetime.datetime,
    Number, Text, Boolean, DateTime, Blank, xlerrors.ExcelError
]]

# However, we are going to define specific types nonetheless, so they can be
# used to cast values to proper internal types.
XlNumber = NewType('XlNumber', _Anything)
XlText = NewType('XlText', _Anything)
XlBoolean = NewType('XlBoolean', _Anything)
XlDateTime = NewType('XlDateTime', _Anything)
XlBlank = NewType('XlBlank', _Anything)
XlArray = NewType('XlArray', Union[_Anything, list, Array])
XlExpr = NewType('XlExpr', Union[_Anything, Expr])

# XlAnything declaration for open-ended return type. XlExpr is specifically
# excluded, since that is not a valid return type.
XlAnything = Union[
    XlNumber, XlText, XlBoolean, XlDateTime, XlBlank, XlArray]



import unittest

from xlcalculator import evaluator, model
from . import testing


class TestEvaluator(unittest.TestCase):

    def setUp(self):
        self.model = model.Model()
        self.model.construct_from_json_file(
            testing.get_resource("model.json"), build_code=True)
        self.evaluator = evaluator.Evaluator(self.model)

    def test_evaluate(self):
        evaluated_result_00 = self.evaluator.evaluate('First!A2')
        result_00 = 0.1
        self.assertEqual(result_00, evaluated_result_00)

        evaluated_result_01 = self.evaluator.evaluate('Seventh!C1')
        result_01 = 3
        self.assertEqual(result_01, evaluated_result_01)

        evaluated_result_02 = self.evaluator.evaluate('Ninth!B1')
        result_02 = 101
        self.assertEqual(result_02, evaluated_result_02)

        evaluated_result_03 = self.evaluator.evaluate('Hundred')
        result_03 = 100
        self.assertEqual(result_03, evaluated_result_03)

        evaluated_result_04 = self.evaluator.evaluate('Tenth!C1')
        result_04 = 102
        self.assertEqual(result_04, evaluated_result_04)

        evaluated_result_06 = self.evaluator.evaluate('Tenth!C2')
        result_06 = 102
        self.assertEqual(result_06, evaluated_result_06)

        evaluated_result_07 = self.evaluator.evaluate('Tenth!C3')
        result_07 = 102
        self.assertEqual(result_07, evaluated_result_07)

    def test_set_value(self):
        self.evaluator.set_cell_value('First!A2', 88)
        evaluated_result_00 = self.evaluator.model.cells['First!A2'].value
        result_00 = 88
        self.assertEqual(result_00, evaluated_result_00)
        # Put it back the way we found it.
        self.evaluator.set_cell_value('First!A2', 0.1)

    def test_set_value_evaluate(self):
        self.evaluator.set_cell_value('First!A2', 88)
        evaluated_result_00 = self.evaluator.evaluate('First!A2')
        result_00 = 88
        self.assertEqual(result_00, evaluated_result_00)
        # Put it back the way we found it.
        self.evaluator.set_cell_value('First!A2', 0.1)

    def test_divide_eval(self):
        div_compiler = model.ModelCompiler()
        div_model = div_compiler.read_and_parse_archive(
            testing.get_resource("division.xlsx"))
        div_evaluator = evaluator.Evaluator(div_model)

        excel_value_00 = div_evaluator.get_cell_value('Sheet1!A1')
        value_00 = div_evaluator.evaluate('Sheet1!A1')
        self.assertEqual(excel_value_00, value_00)

        excel_value_01 = div_evaluator.get_cell_value('Sheet1!B1')
        value_01 = div_evaluator.evaluate('Sheet1!B1')
        self.assertEqual(excel_value_01, value_01)

    def test_subtract_eval(self):
        sub_compiler = model.ModelCompiler()
        sub_model = sub_compiler.read_and_parse_archive(
            testing.get_resource("subtraction.xlsx"))
        sub_evaluator = evaluator.Evaluator(sub_model)

        excel_value_00 = sub_evaluator.get_cell_value('Sheet1!A1')
        value_00 = sub_evaluator.evaluate('Sheet1!A1')
        self.assertEqual(excel_value_00, value_00)

        excel_value_01 = sub_evaluator.get_cell_value('Sheet1!B1')
        value_01 = sub_evaluator.evaluate('Sheet1!B1')
        self.assertEqual(excel_value_01, value_01)

        excel_value_02 = sub_evaluator.get_cell_value('Sheet1!C1')
        value_02 = sub_evaluator.evaluate('Sheet1!C1')
        self.assertEqual(excel_value_02, value_02)

    def test_addition_eval(self):
        add_compiler = model.ModelCompiler()
        add_model = add_compiler.read_and_parse_archive(
            testing.get_resource("addition.xlsx"))
        add_evaluator = evaluator.Evaluator(add_model)

        excel_value_00 = add_evaluator.get_cell_value('Sheet1!A1')
        value_00 = add_evaluator.evaluate('Sheet1!A1')
        self.assertEqual(excel_value_00, value_00)

        excel_value_01 = add_evaluator.get_cell_value('Sheet1!B1')
        value_01 = add_evaluator.evaluate('Sheet1!B1')
        self.assertEqual(excel_value_01, value_01)

        excel_value_02 = add_evaluator.get_cell_value('Sheet1!C1')
        value_02 = add_evaluator.evaluate('Sheet1!C1')
        self.assertEqual(excel_value_02, value_02)

    def test_multiplication_eval(self):
        add_compiler = model.ModelCompiler()
        add_model = add_compiler.read_and_parse_archive(
            testing.get_resource("multiplication.xlsx"))
        add_evaluator = evaluator.Evaluator(add_model)

        excel_value_00 = add_evaluator.get_cell_value('Sheet1!A1')
        value_00 = add_evaluator.evaluate('Sheet1!A1')
        self.assertEqual(excel_value_00, value_00)

        excel_value_01 = add_evaluator.get_cell_value('Sheet1!B1')
        value_01 = add_evaluator.evaluate('Sheet1!B1')
        self.assertEqual(excel_value_01, value_01)

        excel_value_02 = add_evaluator.get_cell_value('Sheet1!C1')
        value_02 = add_evaluator.evaluate('Sheet1!C1')
        self.assertEqual(excel_value_02, value_02)

        excel_value_03 = add_evaluator.get_cell_value('Sheet1!D1')
        value_03 = add_evaluator.evaluate('Sheet1!D1')
        self.assertEqual(excel_value_03, value_03)



from .. import testing


class ExactTest(testing.FunctionalTestCase):
    filename = "EXACT.xlsx"

    def test_evaluation_C2(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!C2')
        value = self.evaluator.evaluate('Sheet1!C2')
        self.assertEqual(excel_value, value)

    def test_evaluation_C3(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!C3')
        value = self.evaluator.evaluate('Sheet1!C3')
        self.assertEqual(excel_value, value)

    def test_evaluation_C4(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!C4')
        value = self.evaluator.evaluate('Sheet1!C4')
        self.assertEqual(excel_value, value)

    def test_evaluation_C6(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!C6')
        value = self.evaluator.evaluate('Sheet1!C6')
        self.assertEqual(excel_value, value)



import unittest

from xlcalculator import xltypes
from xlcalculator.tokenizer import f_token

from . import testing


class XLFormulaTest(testing.XlCalculatorTestCase):

    def test_formula(self):
        self.assertEqual(
            xltypes.XLFormula('=SUM(A1:B1)').formula, '=SUM(A1:B1)')

    def test_tokens(self):
        self.assertASTNodesEqual(
            xltypes.XLFormula('=SUM(A1:B1)').tokens,
            [
                f_token(tvalue='SUM', ttype='function', tsubtype='start'),
                f_token(tvalue='A1:B1', ttype='operand', tsubtype='range'),
                f_token(tvalue='', ttype='function', tsubtype='stop')
            ]
        )

    def test_defined_name_formula(self):
        self.assertASTNodesEqual(
            xltypes.XLFormula('=SUM(A1:B1)').tokens,
            [
                f_token(tvalue='SUM', ttype='function', tsubtype='start'),
                f_token(tvalue='A1:B1', ttype='operand', tsubtype='range'),
                f_token(tvalue='', ttype='function', tsubtype='stop')
            ]
        )


class XLCellTest(unittest.TestCase):

    def test_address(self):
        xlcell = xltypes.XLCell('Sheet1!A1')
        address = 'Sheet1!A1'
        self.assertEqual(address, xlcell.address)

    def test_address_with_range(self):
        with self.assertRaises(Exception) as context:
            xltypes.XLCell('Sheet1!A1:B1')
            self.assertTrue('This is a Range' in context.exception)

    def test_address_with_bad_sheet(self):
        # While the sheet name should be quoted, internally, the code often
        # just puts the sheet name in to produce unique keys, so the utility
        # supports unquoted sheets as well.
        self.assertEqual(
            xltypes.XLCell('Bad Sheet!A1').address, 'Bad Sheet!A1')

    def test_value(self):
        cell = xltypes.XLCell('Sheet1!A1', 5)
        self.assertEqual(cell.value, 5)

    def test_formula(self):
        cell = xltypes.XLCell('Sheet1!A1', 5, 'SUM(A1:B1)')
        self.assertEqual(cell.formula, 'SUM(A1:B1)')

    def test_float(self):
        cell = xltypes.XLCell('Sheet1!A1', 5, 'SUM(A1:B1)')
        self.assertEqual(float(cell), 5.0)

    def test_hash(self):
        cell = xltypes.XLCell('Sheet1!A1', 5, 'SUM(A1:B1)')
        self.assertEqual(hash(cell), hash(('Sheet1', 1, 1)))


class XLRangeTest(unittest.TestCase):

    def test_init(self):
        self.assertEqual(
            xltypes.XLRange('Sheet1!A1:A4').address,
            [['Sheet1!A1'], ['Sheet1!A2'], ['Sheet1!A3'], ['Sheet1!A4']]
        )
        self.assertEqual(
            xltypes.XLRange('Sheet1!A1:A3,Sheet1!C1:C3,Sheet1!E1:E3').address,
            [['Sheet1!A1', 'Sheet1!C1', 'Sheet1!E1'],
             ['Sheet1!A2', 'Sheet1!C2', 'Sheet1!E2'],
             ['Sheet1!A3', 'Sheet1!C3', 'Sheet1!E3']]
        )

    def test_init_with_bad_sheet(self):
        # While the sheet name should be quoted, internally, the code often
        # just puts the sheet name in to produce unique keys, so the utility
        # supports unquoted sheets as well.
        self.assertEqual(
            xltypes.XLRange('Bad Sheet!A1:A3').address,
            [['Bad Sheet!A1'], ['Bad Sheet!A2'], ['Bad Sheet!A3']]
        )

    def test_init_with_multiple_sheet(self):
        with self.assertRaises(ValueError):
            xltypes.XLRange('Sheet1!A1:A4,Sheet2!A1:A4')

    def test_init_with_default_sheet(self):
        self.assertEqual(
            xltypes.XLRange('A1:A3').address,
            [['Sheet1!A1'], ['Sheet1!A2'], ['Sheet1!A3']]
        )

    def test_init_with_spaces(self):
        self.assertEqual(
            xltypes.XLRange('Sheet1!A1:A3, Sheet1!C1:C3').address,
            [['Sheet1!A1', 'Sheet1!C1'],
             ['Sheet1!A2', 'Sheet1!C2'],
             ['Sheet1!A3', 'Sheet1!C3']]
        )

    def test_init_with_one_sheet_spec(self):
        self.assertEqual(
            xltypes.XLRange('Sheet1!A1:A3,C1:C3').address,
            [['Sheet1!A1', 'Sheet1!C1'],
             ['Sheet1!A2', 'Sheet1!C2'],
             ['Sheet1!A3', 'Sheet1!C3']]
        )

    def test_cells(self):
        self.assertEqual(
            xltypes.XLRange('Sheet1!A1:A3').cells,
            [['Sheet1!A1'], ['Sheet1!A2'], ['Sheet1!A3']]
        )



import mock
import typing
import unittest

from xlcalculator.xlfunctions import xl, xlerrors, func_xltypes


class FunctionsTest(unittest.TestCase):

    def test_register(self):
        fn = xl.Functions()

        def sample():
            pass

        fn.register(sample)
        self.assertDictEqual(dict(fn), {'sample': sample})

    def test_register_withName(self):
        fn = xl.Functions()

        def sample():
            pass

        fn.register(sample, 'mysample')
        self.assertDictEqual(dict(fn), {'mysample': sample})

    def test_getattr(self):
        fn = xl.Functions()

        def sample():
            pass

        fn.register(sample)
        self.assertEqual(fn.sample, sample)

    def test_getattr_withUnknown(self):
        fn = xl.Functions()
        with self.assertRaises(AttributeError):
            fn.sample

    def test_register_decorator(self):

        with mock.patch(
            'xlcalculator.xlfunctions.xl.FUNCTIONS', xl.Functions()
        ):

            @xl.register()
            def sample():
                pass

            self.assertDictEqual(dict(xl.FUNCTIONS), {'sample': sample})


class XlModuleTest(unittest.TestCase):

    def test_flatten(self):
        self.assertEqual(xl.flatten([1, [2, 3], [4]]), [1, 2, 3, 4])
        df = func_xltypes.Array([[1, 2], [3, 4]])
        self.assertEqual(xl.flatten(df), [1, 2, 3, 4])
        self.assertEqual(xl.flatten([df]), [1, 2, 3, 4])

    def test_length(self):
        self.assertEqual(xl.length([1, [2, 3], [4]]), 4)
        df = func_xltypes.Array([[1, 2], [3, 4]])
        self.assertEqual(xl.length(df), 4)

    def test_validate_args(self):

        @xl.validate_args
        def func(arg: func_xltypes.XlNumber):
            return arg

        self.assertEqual(func('1'), 1)

    def test_validate_args_fail(self):

        @xl.validate_args
        def func(arg: func_xltypes.XlNumber):
            return arg

        self.assertIsInstance(func('bad'), xlerrors.ValueExcelError)
        self.assertIsInstance(
            func(xlerrors.ValueExcelError()), xlerrors.ValueExcelError)

    def test_validate_args_with_list(self):

        @xl.validate_args
        def func(*args: typing.List[func_xltypes.XlNumber]):
            return args

        self.assertEqual(func('1', 2), (1, 2))

    def test_validate_args_with_union(self):

        @xl.validate_args
        def func(
            arg: typing.Union[func_xltypes.XlNumber, func_xltypes.XlText]
        ):
            return arg

        self.assertEqual(func('data'), 'data')
        self.assertEqual(func(1), 1)

    def test_validate_args_with_union_and_bad_value(self):

        @xl.validate_args
        def func(
            arg: typing.Union[func_xltypes.XlNumber, func_xltypes.XlDateTime]
        ):
            return arg

        self.assertIsInstance(func('data'), xlerrors.ValueExcelError)

    def test_validate_args_without_type(self):

        @xl.validate_args
        def func(arg):
            return arg

        obj = object()
        self.assertEqual(func(obj), obj)



import unittest

from xlcalculator.xlfunctions import xlerrors, func_xltypes, statistics


class StatisticsModuleTest(unittest.TestCase):

    def test_AVERAGE(self):
        self.assertEqual(
            statistics.AVERAGE(func_xltypes.Array([[1, 2], [3, 4]])), 2.5)

    def test_AVERAGE_numbers(self):
        self.assertEqual(statistics.AVERAGE(1, 2.0, 3, 4.0), 2.5)

    def test_AVERAGE_without_any_numbers(self):
        self.assertEqual(statistics.AVERAGE(), 0)

    def test_AVERAGE_mixed(self):
        self.assertEqual(
            statistics.AVERAGE(
                func_xltypes.Array([[1, 2], [3, 4]]), 1, 2, 3, 4
            ),
            2.5
        )

    def test_COUNT(self):
        range0 = func_xltypes.Array([[1, 2], [3, 4]])
        range1 = func_xltypes.Array([[1, 2], [3, 'SPAM']])
        self.assertEqual(statistics.COUNT(range0), 4)
        self.assertEqual(statistics.COUNT(range1), 3)
        self.assertEqual(statistics.COUNT(range0, range1), 7)
        self.assertEqual(statistics.COUNT(range0, range1, 1), 8)
        self.assertEqual(statistics.COUNT(range0, range1, 1, 'SPAM'), 8)

    def test_COUNT_without_any_values(self):
        self.assertIsInstance(
            statistics.COUNT(), xlerrors.ValueExcelError)

    def test_COUNT_with_too_many_values(self):
        self.assertIsInstance(
            statistics.COUNT([0] * 300), xlerrors.ValueExcelError)

    def test_COUNTA(self):
        range0 = func_xltypes.Array([[1, 2], [3, 4]])
        range1 = func_xltypes.Array([[2, 1], [3, '']])
        self.assertEqual(statistics.COUNTA(range0), 4)
        self.assertEqual(statistics.COUNTA(range1), 3)

    def test_COUNTA_with_bad_arg(self):
        self.assertIsInstance(statistics.COUNTA(None), xlerrors.NullExcelError)

    def test_COUNTA_without_any_values(self):
        self.assertIsInstance(statistics.COUNTA(), xlerrors.NullExcelError)

    def test_COUNTA_with_too_many_values(self):
        self.assertIsInstance(
            statistics.COUNTA([0] * 300), xlerrors.ValueExcelError)

    def test_COUNTIF(self):
        countRange = func_xltypes.Array([1, 2, 3, 4])
        condition = ">2"
        self.assertEqual(statistics.COUNTIF(countRange, condition), 2)

    def test_COUNTIF_text(self):
        countRange = func_xltypes.Array(["a", "b", "A"])
        condition = "=A"
        self.assertEqual(statistics.COUNTIF(countRange, condition), 2)

    def test_COUNTIFS(self):
        countRange = func_xltypes.Array([1, 2, 3, 4])
        countRange2 = func_xltypes.Array(["a", "B", "A", "A"])
        condition = "<3"
        condition2 = "A"
        self.assertEqual(statistics.COUNTIFS(
            countRange, condition, countRange2, condition2
        ), 1)

    def test_MAX(self):
        self.assertEqual(
            statistics.MAX(func_xltypes.Array([[1, 2], [3, 4]])),
            4
        )

    def test_MAX_without_any_numbers(self):
        self.assertEqual(statistics.MAX(), 0)

    def test_MIN(self):
        self.assertEqual(
            statistics.MIN(func_xltypes.Array([[1, 2], [3, 4]])),
            1
        )

    def test_MIN_without_any_numbers(self):
        self.assertEqual(statistics.MIN(), 0)

    def test_MIN_with_mixed_types(self):
        self.assertEqual(statistics.MIN(2, 3.0, True), 1)



# ========================================================================
#  Description: Tokenise an Excel formula using an implementation of
#               E. W. Bachtal's algorithm, found here:
#
#                   http://ewbi.blogs.com/develops/2004/12/excel_formula_p.html
#
#               Tested with Python v2.5 (win32)
#       Author: Robin Macharg
#    Copyright: Algorithm (c) E. W. Bachtal, this implementation (c) R. Macharg
#
#  Modification History
#
#  Date         Author Comment
#  =======================================================================
#  2006/11/29 - RMM  - Made strictly class-based.
#                      Added parse, render and pretty print methods
#  2006/11    - RMM  - RMM = Robin Macharg
#                            Created
#  2011/10    - Dirk Gorissen - Patch to support scientific notation
# ========================================================================

import re
from dataclasses import dataclass, field
import uuid
from string import ascii_uppercase


def col2num(col):
    if not col:
        raise Exception("Column may not be empty")

    tot = 0
    for i, c in enumerate([c for c in col[::-1] if c != "$"]):
        if c == '$':
            continue
        tot += (ord(c) - 64) * 26 ** i

    return tot


def num2col(num):
    if num < 1:
        raise Exception("Number must be larger than 0: %s" % num)

    s = ''
    q = num
    while q > 0:
        (q, r) = divmod(q, 26)
        if r == 0:
            q = q - 1
            r = 26
        s = ascii_uppercase[r - 1] + s

    return s


# ========================================================================
#        Class: ExcelParserTokens
#  Description: Inheritable container for token definitions
#
#   Attributes: Self explanatory
#
#      Methods: None
# ========================================================================
class ExcelParserTokens(object):
    TOK_TYPE_NOOP = "noop"
    TOK_TYPE_OPERAND = "operand"
    TOK_TYPE_FUNCTION = "function"
    TOK_TYPE_SUBEXPR = "subexpression"
    TOK_TYPE_ARGUMENT = "argument"
    TOK_TYPE_OP_PRE = "operator-prefix"
    TOK_TYPE_OP_IN = "operator-infix"
    TOK_TYPE_OP_POST = "operator-postfix"
    TOK_TYPE_WSPACE = "white-space"
    TOK_TYPE_UNKNOWN = "unknown"

    TOK_SUBTYPE_START = "start"
    TOK_SUBTYPE_STOP = "stop"
    TOK_SUBTYPE_TEXT = "text"
    TOK_SUBTYPE_NUMBER = "number"
    TOK_SUBTYPE_LOGICAL = "logical"
    TOK_SUBTYPE_ERROR = "error"
    TOK_SUBTYPE_RANGE = "range"
    TOK_SUBTYPE_MATH = "math"
    TOK_SUBTYPE_CONCAT = "concatenate"
    TOK_SUBTYPE_INTERSECT = "intersect"
    TOK_SUBTYPE_UNION = "union"
    TOK_SUBTYPE_NONE = "none"


def init_uuid():
    """Default factory to initialise Formula.ranges."""
    return uuid.uuid4()


# ========================================================================
#        Class: f_token
#  Description: Encapsulate a formula token
#
#   Attributes:   tvalue -
#                  ttype - See token definitions, above, for values
#               tsubtype - See token definitions, above, for values
#
#      Methods: f_token  - __init__()
# ========================================================================
@dataclass
class f_token:

    tvalue: str
    ttype: str
    tsubtype: str
    unique_identifier: uuid = field(
        init=False, default_factory=init_uuid, compare=True, hash=True,
        repr=True)

    def __repr__(self):
        return "<{} tvalue: {} ttype: {} tsubtype: {}>".format(
            self.__class__.__name__, self.tvalue, self.ttype, self.tsubtype)

    def __str__(self):
        return self.__repr__()


# ========================================================================
#        Class: f_tokens
#  Description: An ordered list of tokens
#
#   Attributes:        items - Ordered list
#                      index - Current position in the list
#
#      Methods: f_tokens     - __init__()
#               f_token      - add()      - Add a token to the end of the list
#               None         - addRef()   - Add a token to the end of the list
#               None         - reset()    - reset the index to -1
#               Boolean      - BOF()      - End of list?
#               Boolean      - EOF()      - Beginning of list?
#               Boolean      - moveNext() - Move the index along one
#               f_token/None - current()  - Return the current token
#               f_token/None - next()     - Return the next token (leave the
#                                           index unchanged)
#               f_token/None - previous() - Return the previous token (leave
#                                           the index unchanged)
# ========================================================================
class f_tokens(object):

    def __init__(self):
        self.items = []
        self.index = -1

    def add(self, value, type, subtype=""):
        if (not subtype):
            subtype = ""
        token = f_token(value, type, subtype)
        self.addRef(token)
        return token

    def addRef(self, token):
        self.items.append(token)

    def reset(self):
        self.index = -1

    def BOF(self):
        return self.index <= 0

    def EOF(self):
        return self.index >= (len(self.items) - 1)

    def moveNext(self):
        if self.EOF():
            return False
        self.index += 1
        return True

    def current(self):
        if self.index == -1:
            return None
        return self.items[self.index]

    def __next__(self):
        if self.EOF():
            return None
        return self.items[self.index + 1]

    def previous(self):
        if self.index < 1:
            return None
        return self.items[self.index - 1]

    # Make this object pass as an iterator.
    def __iter__(self):
        self.reset()
        return self

    def next(self):
        return self.__next__()


# ========================================================================
#        Class: f_tokenStack
#     Inherits: ExcelParserTokens - a list of token values
#  Description: A LIFO stack of tokens
#
#   Attributes:        items - Ordered list
#
#      Methods: f_tokenStack - __init__()
#               None         - push(token) - Push a token onto the stack
#               f_token/None - pop()       - Pop a token off the stack
#               f_token/None - token()     - Non-destructively return the top
#                                            item on the stack
#               String       - type()      - Return the top token's type
#               String       - subtype()   - Return the top token's subtype
#               String       - value()     - Return the top token's value
# ========================================================================
class f_tokenStack(ExcelParserTokens):

    def __init__(self):
        self.items = []

    def push(self, token):
        self.items.append(token)

    def pop(self):
        token = self.items.pop()
        return f_token("", token.ttype, self.TOK_SUBTYPE_STOP)

    def token(self):
        # Note: this uses Pythons and/or "hack" to emulate C's ternary
        # operator (i.e. cond ? exp1 : exp2)
        return (
            (
                (len(self.items) > 0)
                and [self.items[len(self.items) - 1]]
                or [None]
            )[0]
        )

    def value(self):
        return ((self.token()) and [(self.token()).tvalue] or [""])[0]

    def type(self):
        return ((self.token()) and [(self.token()).ttype] or [""])[0]

    def subtype(self):
        return ((self.token()) and [(self.token()).tsubtype] or [""])[0]


# ========================================================================
#        Class: ExcelParser
#  Description: Parse an Excel formula into a stream of tokens
#
#   Attributes:
#
#      Methods: f_tokens - getTokens(formula) - return a token stream (list)
# ========================================================================
class ExcelParser(ExcelParserTokens):

    def __init__(self, tokenize_range=False):
        if tokenize_range:
            self.OPERATORS = "+-*/^&=><:"
        else:
            self.OPERATORS = "+-*/^&=><"

    def getTokens(self, formula):

        def currentChar():
            return formula[offset]

        def doubleChar():
            return formula[offset:offset + 2]

        def nextChar():
            # JavaScript returns an empty string if the index is out of bounds,
            # Python throws an IndexError.  We mimic this behaviour here.
            try:
                formula[offset + 1]
            except IndexError:
                return ""
            else:
                return formula[offset + 1]

        def EOF():
            return offset >= len(formula)

        tokens = f_tokens()
        tokenStack = f_tokenStack()
        offset = 0
        token = ""
        inString = False
        inPath = False
        inRange = False
        inError = False

        while (len(formula) > 0):
            if (formula[0] in (" ", "\n")):
                formula = formula[1:]

            else:
                if (formula[0] == "="):
                    formula = formula[1:]
                break

        # state-dependent character evaluation (order is important)
        while not EOF():
            # double-quoted strings
            # embeds are doubled
            # end marks token
            if inString:
                if currentChar() == "\"":
                    if nextChar() == "\"":
                        token += "\""
                        offset += 1

                    else:
                        inString = False
                        tokens.add(
                            token, self.TOK_TYPE_OPERAND,
                            self.TOK_SUBTYPE_TEXT)
                        token = ""

                else:
                    token += currentChar()
                offset += 1
                continue

            # single-quoted strings (links)
            # embeds are double
            # end does not mark a token
            if inPath:
                if currentChar() == "'":
                    if nextChar() == "'":
                        token += "'"
                        offset += 1

                    else:
                        inPath = False

                else:
                    token += currentChar()
                offset += 1
                continue

            # bracketed strings (range offset or linked workbook name)
            # no embeds (changed to "()" by Excel)
            # end does not mark a token
            if inRange:
                if currentChar() == "]":
                    inRange = False
                token += currentChar()
                offset += 1
                continue

            # error values
            # end marks a token, determined from absolute list of values
            if inError:
                token += currentChar()
                offset += 1
                if ",#NULL!,#DIV/0!,#VALUE!,#REF!,#NAME?,#NUM!,#N/A,".find(
                        "," + token + ",") != -1:
                    inError = False
                    tokens.add(
                        token, self.TOK_TYPE_OPERAND, self.TOK_SUBTYPE_ERROR)
                    token = ""
                continue

            # scientific notation check
            regexSN = r'^[1-9]{1}(\.[0-9]+)?[eE]{1}$'
            if (("+-").find(currentChar()) != -1):
                if len(token) > 1:
                    if re.match(regexSN, token):
                        token += currentChar()
                        offset += 1
                        continue

            # independent character evaulation (order not important)
            #
            # establish state-dependent character evaluations
            if currentChar() == "\"":
                if len(token) > 0:
                    # not expected
                    tokens.add(token, self.TOK_TYPE_UNKNOWN)
                    token = ""
                inString = True
                offset += 1
                continue

            if currentChar() == "'":
                if len(token) > 0:
                    # not expected
                    tokens.add(token, self.TOK_TYPE_UNKNOWN)
                    token = ""
                inPath = True
                offset += 1
                continue

            if (currentChar() == "["):
                inRange = True
                token += currentChar()
                offset += 1
                continue

            if (currentChar() == "#"):
                if (len(token) > 0):
                    # not expected
                    tokens.add(token, self.TOK_TYPE_UNKNOWN)
                    token = ""
                inError = True
                token += currentChar()
                offset += 1
                continue

            # mark start and end of arrays and array rows
            if (currentChar() == "{"):
                if (len(token) > 0):
                    # not expected
                    tokens.add(token, self.TOK_TYPE_UNKNOWN)
                    token = ""
                tokenStack.push(tokens.add(
                    "ARRAY",
                    self.TOK_TYPE_FUNCTION, self.TOK_SUBTYPE_START))
                tokenStack.push(tokens.add(
                    "ARRAYROW",
                    self.TOK_TYPE_FUNCTION, self.TOK_SUBTYPE_START))
                offset += 1
                continue

            if (currentChar() == ";"):
                if (len(token) > 0):
                    tokens.add(token, self.TOK_TYPE_OPERAND)
                    token = ""
                tokens.addRef(tokenStack.pop())
                tokens.add(",", self.TOK_TYPE_ARGUMENT)
                tokenStack.push(tokens.add(
                    "ARRAYROW",
                    self.TOK_TYPE_FUNCTION, self.TOK_SUBTYPE_START))
                offset += 1
                continue

            if (currentChar() == "}"):
                if (len(token) > 0):
                    tokens.add(token, self.TOK_TYPE_OPERAND)
                    token = ""
                tokens.addRef(tokenStack.pop())
                tokens.addRef(tokenStack.pop())
                offset += 1
                continue

            # trim white-space
            if (currentChar() in (" ", "\n")):
                if (len(token) > 0):
                    tokens.add(token, self.TOK_TYPE_OPERAND)
                    token = ""
                tokens.add("", self.TOK_TYPE_WSPACE)
                offset += 1
                while ((currentChar() in (" ", "\n")) and (not EOF())):
                    offset += 1
                continue

            # multi-character comparators
            if (",>=,<=,<>,".find("," + doubleChar() + ",") != -1):
                if (len(token) > 0):
                    tokens.add(token, self.TOK_TYPE_OPERAND)
                    token = ""
                tokens.add(
                    doubleChar(),
                    self.TOK_TYPE_OP_IN, self.TOK_SUBTYPE_LOGICAL)
                offset += 2
                continue

            # standard infix operators
            if (self.OPERATORS.find(currentChar()) != -1):
                if (len(token) > 0):
                    tokens.add(token, self.TOK_TYPE_OPERAND)
                    token = ""
                tokens.add(currentChar(), self.TOK_TYPE_OP_IN)
                offset += 1
                continue

            # standard postfix operators
            if ("%".find(currentChar()) != -1):
                if (len(token) > 0):
                    tokens.add(float(token) / 100, self.TOK_TYPE_OPERAND)
                    token = ""
                else:
                    tokens.add('*', self.TOK_TYPE_OP_IN)
                    tokens.add(0.01, self.TOK_TYPE_OPERAND)
                # tokens.add(currentChar(), self.TOK_TYPE_OP_POST)
                offset += 1
                continue

            # start subexpression or function
            if (currentChar() == "("):
                if (len(token) > 0):
                    tokenStack.push(tokens.add(
                        token, self.TOK_TYPE_FUNCTION, self.TOK_SUBTYPE_START))
                    token = ""
                else:
                    tokenStack.push(tokens.add(
                        "", self.TOK_TYPE_SUBEXPR, self.TOK_SUBTYPE_START))
                offset += 1
                continue

            # function, subexpression, array parameters
            if (currentChar() == ","):
                if (len(token) > 0):
                    tokens.add(token, self.TOK_TYPE_OPERAND)
                    token = ""
                if (not (tokenStack.type() == self.TOK_TYPE_FUNCTION)):
                    tokens.add(
                        currentChar(),
                        self.TOK_TYPE_OP_IN, self.TOK_SUBTYPE_UNION)
                else:
                    tokens.add(currentChar(), self.TOK_TYPE_ARGUMENT)
                offset += 1
                if (currentChar() == ","):
                    tokens.add(
                        'None',
                        self.TOK_TYPE_OPERAND, self.TOK_SUBTYPE_NONE)
                    token = ""
                continue

            # stop subexpression
            if (currentChar() == ")"):
                if (len(token) > 0):
                    tokens.add(token, self.TOK_TYPE_OPERAND)
                    token = ""
                tokens.addRef(tokenStack.pop())
                offset += 1
                continue

            # token accumulation
            token += currentChar()
            offset += 1

        # dump remaining accumulation
        if (len(token) > 0):
            tokens.add(token, self.TOK_TYPE_OPERAND)

        # move all tokens to a new collection, excluding all unnecessary
        # white-space tokens
        tokens2 = f_tokens()

        while (tokens.moveNext()):
            token = tokens.current()

            if (token.ttype == self.TOK_TYPE_WSPACE):
                if ((tokens.BOF()) or (tokens.EOF())):
                    pass
                elif (not(
                     (
                         (tokens.previous().ttype == self.TOK_TYPE_FUNCTION)
                         and (tokens.previous().tsubtype
                              == self.TOK_SUBTYPE_STOP)
                     ) or (
                         (tokens.previous().ttype == self.TOK_TYPE_SUBEXPR)
                         and (tokens.previous().tsubtype
                              == self.TOK_SUBTYPE_STOP)
                     ) or (
                         tokens.previous().ttype == self.TOK_TYPE_OPERAND
                     )
                )):
                    pass
                elif (not(
                    (
                        (tokens.next().ttype == self.TOK_TYPE_FUNCTION)
                        and (tokens.next().tsubtype
                             == self.TOK_SUBTYPE_START)
                    ) or (
                        (tokens.next().ttype == self.TOK_TYPE_SUBEXPR)
                        and (tokens.next().tsubtype == self.TOK_SUBTYPE_START)
                    ) or (
                        tokens.next().ttype == self.TOK_TYPE_OPERAND)
                )):
                    pass
                else:
                    tokens2.add(
                        token.tvalue, self.TOK_TYPE_OP_IN,
                        self.TOK_SUBTYPE_INTERSECT)
                continue

            tokens2.addRef(token)

        # switch infix "-" operator to prefix when appropriate, switch infix
        # "+" operator to noop when appropriate, identify operand and
        # infix-operator subtypes, pull "@" from in front of function names
        while (tokens2.moveNext()):
            token = tokens2.current()
            if (
                    (token.ttype == self.TOK_TYPE_OP_IN)
                    and (token.tvalue == "-")
            ):
                if (tokens2.BOF()):
                    token.ttype = self.TOK_TYPE_OP_PRE
                elif (
                    (
                        (tokens2.previous().ttype == self.TOK_TYPE_FUNCTION)
                        and (tokens2.previous().tsubtype
                             == self.TOK_SUBTYPE_STOP)
                    ) or (
                        (tokens2.previous().ttype == self.TOK_TYPE_SUBEXPR)
                        and (tokens2.previous().tsubtype
                             == self.TOK_SUBTYPE_STOP)
                    ) or (
                        tokens2.previous().ttype == self.TOK_TYPE_OP_POST
                    ) or (
                        tokens2.previous().ttype == self.TOK_TYPE_OPERAND
                    )
                ):
                    token.tsubtype = self.TOK_SUBTYPE_MATH

                else:
                    token.ttype = self.TOK_TYPE_OP_PRE

                continue

            if (
                    (token.ttype == self.TOK_TYPE_OP_IN)
                    and (token.tvalue == "+")
            ):
                if tokens2.BOF():
                    token.ttype = self.TOK_TYPE_NOOP
                elif (
                    (
                        (tokens2.previous().ttype == self.TOK_TYPE_FUNCTION)
                        and (tokens2.previous().tsubtype
                             == self.TOK_SUBTYPE_STOP)
                    ) or (
                        (tokens2.previous().ttype == self.TOK_TYPE_SUBEXPR)
                        and (tokens2.previous().tsubtype
                             == self.TOK_SUBTYPE_STOP)
                    ) or (
                        tokens2.previous().ttype == self.TOK_TYPE_OP_POST
                    ) or (
                        tokens2.previous().ttype == self.TOK_TYPE_OPERAND
                    )
                ):
                    token.tsubtype = self.TOK_SUBTYPE_MATH

                else:
                    token.ttype = self.TOK_TYPE_NOOP

                continue

            if ((token.ttype == self.TOK_TYPE_OP_IN)
                    and (len(token.tsubtype) == 0)):
                if (("<>=").find(token.tvalue[0:1]) != -1):
                    token.tsubtype = self.TOK_SUBTYPE_LOGICAL

                elif (token.tvalue == "&"):
                    token.tsubtype = self.TOK_SUBTYPE_CONCAT

                else:
                    token.tsubtype = self.TOK_SUBTYPE_MATH

                continue

            if ((token.ttype == self.TOK_TYPE_OPERAND)
                    and (len(token.tsubtype) == 0)):
                try:
                    float(token.tvalue)

                except ValueError:
                    if ((token.tvalue == 'TRUE') or (token.tvalue == 'FALSE')):
                        token.tsubtype = self.TOK_SUBTYPE_LOGICAL
                    else:
                        token.tsubtype = self.TOK_SUBTYPE_RANGE
                else:
                    token.tsubtype = self.TOK_SUBTYPE_NUMBER

                continue

            if (token.ttype == self.TOK_TYPE_FUNCTION):
                if (token.tvalue[0:1] == "@"):
                    token.tvalue = token.tvalue[1:]

                continue

        tokens2.reset()

        # move all tokens to a new collection, excluding all noops
        tokens = f_tokens()
        while (tokens2.moveNext()):
            if (tokens2.current().ttype != self.TOK_TYPE_NOOP):
                tokens.addRef(tokens2.current())

        tokens.reset()
        return tokens

    def parse(self, formula):
        self.tokens = self.getTokens(formula)
        return self.tokens



from .. import testing


class DoubleMinusTest(testing.FunctionalTestCase):
    filename = "double_minus.xlsx"

    def test_evaluation_double_unary_A_eq_B(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A1')
        value = self.evaluator.evaluate('Sheet1!A1')
        self.assertEqual(excel_value, value)

    def test_evaluation_double_unary_number(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A2')
        value = self.evaluator.evaluate('Sheet1!A2')
        self.assertEqual(excel_value, value)

    def test_evaluation_double_unary_formula(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A3')
        value = self.evaluator.evaluate('Sheet1!A3')
        self.assertEqual(excel_value, value)



from . import xl, xlerrors, func_xltypes


@xl.register()
@xl.validate_args
def ISBLANK(cell: func_xltypes.XlAnything) -> func_xltypes.Boolean:
    """Returns TRUE if the cell is empty.
    """
    return isinstance(cell, func_xltypes.Blank) or cell.value == ''


@xl.register()
def ISERR(value: func_xltypes.XlAnything) -> func_xltypes.Boolean:
    """Value refers to error values
    (#VALUE!, #REF!, #DIV/0!, #NUM!, #NAME?, or #NULL!)
    And NOT error #N/A
    """
    if isinstance(value, xlerrors.ExcelError) \
            and not isinstance(value, xlerrors.NaExcelError):
        return True
    else:
        return False


@xl.register()
def ISERROR(value: func_xltypes.XlAnything) -> func_xltypes.Boolean:
    """Value refers to any error value
    (#N/A, #VALUE!, #REF!, #DIV/0!, #NUM!, #NAME?, or #NULL!).
    """
    return isinstance(value, xlerrors.ExcelError)


@xl.register()
@xl.validate_args
def ISEVEN(num: func_xltypes.XlNumber) -> func_xltypes.Boolean:
    """Returns TRUE if number is even, or FALSE if number is odd.

    https://support.microsoft.com/en-us/office/
        iseven-function-aa15929a-d77b-4fbb-92f4-2f479af55356
    """

    if int(num) == 1:
        return False

    elif (int(num) % 2) == 0:
        return True

    else:
        return False


@xl.register()
@xl.validate_args
def ISNUMBER(cell: func_xltypes.XlAnything) -> func_xltypes.Boolean:
    """Returns True if the cell is number.
    """
    return isinstance(cell, func_xltypes.Number)


@xl.register()
@xl.validate_args
def ISTEXT(cell: func_xltypes.XlAnything) -> func_xltypes.Boolean:
    """Returns True if the cell is text.
    """
    return isinstance(cell, func_xltypes.Text)


@xl.register()
@xl.validate_args
def NA() -> xlerrors.ExcelError:
    return xlerrors.NaExcelError()


@xl.register()
def ISNA(cell) -> func_xltypes.Boolean:
    """Returns True if the cell is #N/A.

    Don't call validate_args here because we allow errors to be
    passed in.

    """
    return isinstance(cell, xlerrors.NaExcelError)


@xl.register()
@xl.validate_args
def ISODD(num: func_xltypes.XlNumber) -> func_xltypes.Boolean:
    """Returns TRUE if number is odd, or FALSE if number is even.

    https://support.microsoft.com/en-us/office/
        isodd-function-1208a56d-4f10-4f44-a5fc-648cafd6c07a
    """

    if int(num) == 1:
        return True

    elif (int(num) % 2) == 0:
        return False

    else:
        return True




from typing import Tuple

import pandas as pd
import numpy_financial as npf
from scipy.optimize import newton

from . import xl, xlerrors, func_xltypes


@xl.register()
@xl.validate_args
def IRR(
        values: func_xltypes.XlArray,
        guess: func_xltypes.XlNumber = None
) -> func_xltypes.XlNumber:
    """Returns the internal rate of return for a series of cash flows

    https://support.office.com/en-us/article/
        irr-function-64925eaa-9988-495b-b290-3ad0c163c1bc
    """
    # `guess` is not used, but unnecessary, since it is a pure perforamnce
    # optimization.
    return npf.irr(xl.flatten(values))


@xl.register()
@xl.validate_args
def NPV(
        rate: func_xltypes.XlNumber,
        *values: Tuple[func_xltypes.XlNumber],
) -> func_xltypes.XlNumber:
    """Calculates the net present value of an investment by using a discount
    rate and a series of future payments (negative values) and income
    (positive values).

    https://support.office.com/en-us/article/
        npv-function-8672cb67-2576-4d07-b67b-ac28acf2a568
    """
    if not len(values):
        raise xlerrors.ValueExcelError('value1 is required')

    cashflow = [float(value) for value in values]
    rate = float(rate)

    if xl.COMPATIBILITY == 'PYTHON':
        return npf.npv(rate, cashflow)

    return sum([
        val * (1 + rate) ** - (i + 1)
        for (i, val) in enumerate(cashflow)
    ])


@xl.register()
def PMT(
        rate: func_xltypes.XlNumber,
        nper: func_xltypes.XlNumber,
        pv: func_xltypes.XlNumber,
        fv: func_xltypes.XlNumber = 0,
        type: func_xltypes.XlNumber = 0
) -> func_xltypes.XlNumber:
    """Calculates the payment for a loan based on constant payments and
    a constant interest rate.

    https://support.office.com/en-us/article/
        pmt-function-0214da64-9a63-4996-bc20-214433fa6441
    """
    # WARNING fv & type not used yet - both are assumed to be their
    #         defaults (0)
    # fv = args[3]
    # type = args[4]

    if xl.COMPATIBILITY == 'PYTHON':
        when = 'end'
        if type != 0:
            when = 'begin'
        return float(npf.pmt(
            float(rate), float(nper), float(pv), fv=float(fv), when=when))

    # return -pv * rate / (1 - power(1 + rate, -nper))
    return float(npf.pmt(
        float(rate), float(nper), float(pv), fv=float(fv), when='end'))


@xl.register()
@xl.validate_args
def PV(
        rate: func_xltypes.XlNumber,
        nper: func_xltypes.XlNumber,
        pmt: func_xltypes.XlNumber,
        fv: func_xltypes.XlNumber = 0,
        type: func_xltypes.XlNumber = 0
) -> func_xltypes.XlNumber:
    """PV, one of the financial functions, calculates the present value of a
    loan or an investment, based on a constant interest rate.

    https://support.office.com/en-us/article/
        pv-function-23879d31-0e02-4321-be01-da16e8168cbd
    """

    return npf.pv(float(rate),
                  float(nper),
                  float(pmt),
                  fv=float(fv),
                  when=int(type))


@xl.register()
def SLN(
        cost: func_xltypes.XlNumber,
        salvage: func_xltypes.XlNumber,
        life: func_xltypes.XlNumber
) -> func_xltypes.XlNumber:
    """Returns the straight-line depreciation of an asset for one period.

    https://support.office.com/en-us/article/
        sln-function-cdb666e5-c1c6-40a7-806a-e695edc2f1c8
    """
    return (cost - salvage) / life


@xl.register()
@xl.validate_args
def VDB(
        cost: func_xltypes.XlNumber,
        salvage: func_xltypes.XlNumber,
        life: func_xltypes.XlNumber,
        start_period: func_xltypes.XlNumber,
        end_period: func_xltypes.XlNumber,
        factor: func_xltypes.XlNumber = 2,
        no_switch: func_xltypes.XlBoolean = False
) -> func_xltypes.XlNumber:
    """Returns the depreciation of an asset for any period you specify.

    https://support.office.com/en-us/article/
        vdb-function-dde4e207-f3fa-488d-91d2-66d55e861d73
    """
    start_period = start_period
    end_period = end_period
    sln_depr = SLN(cost, salvage, life)
    depr_rate = factor / life
    acc_depr = 0
    depr = 0
    switch_to_sln = False
    sln_depr = 0
    result = 0
    start_life = 0
    # Handle cases when life is not an integer.
    if life % 1 > 0:
        end_life = int(life + 1)

    else:
        end_life = int(life)
    periods = list(range(start_life, end_life))

    if int(start_period) != start_period:
        delta_start = abs(int(start_period) - start_period)
        depr = (cost - acc_depr) * depr_rate * delta_start
        acc_depr += depr
        start_life = 1
        periods = [x + 0.5 for x in periods]

    for index, current_year in enumerate(periods):

        if not no_switch:  # no_switch = False (Default Case)
            if switch_to_sln:
                depr = sln_depr

            else:
                depr = (cost - acc_depr) * depr_rate
                acc_depr += depr

                temp_sln_depr = SLN(cost, salvage, life)

                if depr < temp_sln_depr:
                    switch_to_sln = True
                    fixed_remaining_years = life - current_year - 1
                    fixed_remaining_cost = cost - acc_depr

                    # We need to check future sln: current depr should never
                    # be smaller than sln to come.
                    sln_depr = SLN(
                        fixed_remaining_cost, salvage, fixed_remaining_years)

                    # if it's the case, we switch to sln earlier than the
                    # regular case
                    if sln_depr > depr:
                        # cancel what has been done
                        acc_depr -= depr
                        fixed_remaining_years += 1
                        fixed_remaining_cost = cost - acc_depr

                        # recalculate depreciation
                        sln_depr = SLN(
                            fixed_remaining_cost, salvage,
                            fixed_remaining_years)
                        depr = sln_depr
                        acc_depr += depr

        else:  # no_switch = True
            depr = (cost - acc_depr) * depr_rate
            acc_depr += depr

        delta_start = abs(current_year - start_period)

        if delta_start < 1 and delta_start != 0:
            result += depr * (1 - delta_start)

        elif current_year >= start_period and current_year < end_period:

            delta_end = abs(end_period - current_year)

            if delta_end < 1 and delta_end != 0:
                result += depr * delta_end

            else:
                result += depr

    return result


def _xnpv(rate, values, dates):
    if rate <= -1.0:
        return float('inf')

    return sum([value / ((1.0 + rate)**((date - dates[0]) / 365))
               for value, date in zip(values, dates)])


def _xirr(values, dates, guess=None):
    try:
        return newton(lambda r: _xnpv(r, values, dates), guess, maxiter=100)

    except (RuntimeError, FloatingPointError):
        raise xlerrors.NumExcelError('XIRR did not converge')


@xl.register()
@xl.validate_args
def XIRR(
        values: func_xltypes.XlArray,
        dates: func_xltypes.XlArray,
        guess: func_xltypes.XlNumber = 0.1
) -> func_xltypes.XlNumber:
    """Returns the internal rate of return for a schedule of cash flows that
        is not necessarily periodic.

    https://support.microsoft.com/en-us/office/
        xirr-function-de1242ec-6477-445b-b11b-a303ad9adc9d

    Algorithm found on stackoverflow:
    https://stackoverflow.com/questions/63797804/
        python-irr-function-giving-different-result-than-excel-xirr

    From MS, Newton's method is used to optimize:
    https://docs.microsoft.com/en-us/office/troubleshoot/excel/
        algorithm-of-xirr-funcation
    """

    values = values.flatten(func_xltypes.Number, None)
    dates = dates.flatten(func_xltypes.DateTime, None)
    # need to cast dates and guess to Python types else optimizer complains
    dates = [float(date) for date in dates]
    guess = float(guess)

    # TODO: Ignore non numeric cells and boolean cells.
    if len(values) != len(dates):
        raise xlerrors.NumExcelError(
            f'`values` range must be the same length as `dates` range '
            f'in XIRR, {len(values)} != {len(dates)}')

    series = pd.DataFrame({"dates": dates, "values": values})

    # Filter all rows with 0 cashflows
    series = series[series['values'] != 0]

    # Sort dataframe by date
    series = series.sort_values('dates', ascending=True)
    series['values'] = series['values'].astype('float')

    # Create separate lists for values and dates
    series_values = list(series['values'])
    series_dates = list(series['dates'])

    # Calculate IRR
    return _xirr(series_values, series_dates, guess)


@xl.register()
@xl.validate_args
def XNPV(
        rate: func_xltypes.XlNumber,
        values: func_xltypes.XlArray,
        dates: func_xltypes.XlArray,
) -> func_xltypes.XlNumber:
    """Returns the net present value for a schedule of cash flows that
    is not necessarily periodic.

    https://support.microsoft.com/en-us/office/
        xnpv-function-1b42bbf6-370f-4532-a0eb-d67c16b664b7
    """
    values = values.flatten(func_xltypes.Number, None)
    dates = dates.flatten(func_xltypes.DateTime, None)

    # TODO: Ignore non numeric cells and boolean cells.
    if len(values) != len(dates):
        raise xlerrors.NumExcelError(
            f'`values` range must be the same length as `dates` range '
            f'in XNPV, {len(values)} != {len(dates)}')

    return _xnpv(rate, values, dates)



import unittest

from xlcalculator.xlfunctions import information, xlerrors, func_xltypes, date


class InformationModuleTest(unittest.TestCase):

    def test_ISBLANK(self):
        self.assertTrue(information.ISBLANK(func_xltypes.Blank()))
        self.assertFalse(information.ISBLANK(func_xltypes.Text("hello")))
        self.assertTrue(information.ISBLANK(func_xltypes.Text("")))

    def test_ISTEXT(self):
        self.assertTrue(information.ISTEXT(func_xltypes.Text("hello")))
        self.assertFalse(information.ISTEXT(func_xltypes.Number(1234)))
        self.assertTrue(information.ISTEXT("hello"))
        self.assertFalse(information.ISTEXT(1234))
        self.assertTrue(information.ISTEXT("1234"))

    def test_NA(self):
        self.assertEqual(information.NA(), xlerrors.ERROR_CODE_NA)

    def test_ISNA(self):
        self.assertTrue(information.ISNA(information.NA()))
        self.assertFalse(information.ISNA(xlerrors.ValueExcelError))

    def test_ISEVEN(self):
        self.assertFalse(information.ISEVEN(func_xltypes.Number(1)))
        self.assertTrue(information.ISEVEN(func_xltypes.Number(2)))

        self.assertFalse(information.ISEVEN(-1))
        self.assertTrue(information.ISEVEN(2.5))
        self.assertFalse(information.ISEVEN(5))
        self.assertTrue(information.ISEVEN(0))
        self.assertTrue(information.ISEVEN(date.DATE(2011, 12, 23)))

    def test_ISEVEN_error(self):
        self.assertIsInstance(information.ISEVEN(func_xltypes.Text("hello")),
                              xlerrors.ValueExcelError)

    def test_ISODD(self):
        self.assertEqual(information.ISODD(1), True)
        self.assertEqual(information.ISODD(2), False)

        self.assertEqual(information.ISODD(-1), True)
        self.assertEqual(information.ISODD(2.5), False)
        self.assertEqual(information.ISODD(5), True)

    def test_ISODD_error(self):
        self.assertIsInstance(information.ISODD(func_xltypes.Text("hello")),
                              xlerrors.ValueExcelError)

    def test_ISNUMBER(self):
        self.assertFalse(information.ISNUMBER(func_xltypes.Text("hello")))
        self.assertTrue(information.ISNUMBER(func_xltypes.Number(1234)))
        self.assertFalse(information.ISNUMBER("hello"))
        self.assertTrue(information.ISNUMBER(1234))
        self.assertFalse(information.ISNUMBER("1234"))

    def test_ISERR(self):
        num_error = xlerrors.NumExcelError()
        na_error = xlerrors.NaExcelError()

        self.assertTrue(information.ISERR(num_error))
        self.assertFalse(information.ISERR(na_error))

    def test_ISERROR(self):
        num_error = xlerrors.NumExcelError()
        na_error = xlerrors.NaExcelError()

        self.assertTrue(information.ISERROR(num_error))
        self.assertTrue(information.ISERROR(na_error))



from .. import testing


class CountIfsTest(testing.FunctionalTestCase):
    filename = "COUNTIFS.xlsx"

    def test_evaluation_A10(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A10')
        value = self.evaluator.evaluate('Sheet1!A10')
        self.assertEqual(excel_value, value)



from typing import Tuple

from . import xl, xlerrors, func_xltypes


@xl.register()
@xl.validate_args
def AND(
        *logicals: Tuple[func_xltypes.XlExpr]
) -> func_xltypes.XlBoolean:
    """Determine if all conditions in a test are TRUE

    https://support.office.com/en-us/article/
        and-function-5f19b2e8-e1df-4408-897a-ce285a19e9d9
    """
    if not logicals:
        raise xlerrors.NullExcelError('logical1 is required')

    # Use delayed evaluation to minimize th amount of valaues to evaluate.
    for logical in logicals:
        val = logical()
        for item in xl.flatten([val]):
            if func_xltypes.Blank.is_blank(item):
                continue
            if not bool(item):
                return False

    return True


@xl.register()
@xl.validate_args
def FALSE() -> func_xltypes.XlBoolean:
    """Returns the logical value FALSE.

    https://support.office.com/en-us/article/
        false-function-2d58dfa5-9c03-4259-bf8f-f0ae14346904
    """
    return False


@xl.register()
@xl.validate_args
def OR(
        *logicals: Tuple[func_xltypes.XlExpr]
) -> func_xltypes.XlBoolean:
    """Determine if any conditions in a test are TRUE.

    https://support.office.com/en-us/article/
        or-function-7d17ad14-8700-4281-b308-00b131e22af0
    """
    if not logicals:
        raise xlerrors.NullExcelError('logical1 is required')

    # Use delayed evaluation to minimize th amount of valaues to evaluate.
    for logical in logicals:
        val = logical()
        for item in xl.flatten([val]):
            if func_xltypes.Blank.is_blank(item):
                continue
            if bool(item):
                return True

    return False


@xl.register()
@xl.validate_args
def IF(
        logical_test: func_xltypes.XlExpr,
        value_if_true: func_xltypes.XlExpr = True,
        value_if_false: func_xltypes.XlExpr = False
):
    """Return one value if a condition is true and another value if it's false.

    https://support.office.com/en-us/article/
        if-function-69aed7c9-4e8a-4755-a9bc-aa8bbff73be2
    """
    # Use delayed evaluation to only evaluate the true or false value but not
    # both.
    return value_if_true() if logical_test() else value_if_false()


@xl.register()
@xl.validate_args
def TRUE() -> func_xltypes.XlBoolean:
    """Returns the logical value TRUE.

    https://support.office.com/en-us/article/
        true-function-7652c6e3-8987-48d0-97cd-ef223246b3fb
    """
    return True



from .. import testing


class VLookupTest(testing.FunctionalTestCase):
    filename = "VLOOKUP.xlsx"

    def test_evaluation_B7(self):
        # Range Match - Unsupposrted
        with self.assertRaises(RuntimeError):
            self.evaluator.evaluate('Sheet1!B7')

    def test_evaluation_E7(self):
        # Exact Match.
        excel_value = self.evaluator.get_cell_value('Sheet1!E7')
        value = self.evaluator.evaluate('Sheet1!E7')
        self.assertEqual(excel_value, value)



import unittest

from xlcalculator.xlfunctions import lookup, xlerrors, func_xltypes


class LookupModuleTest(unittest.TestCase):

    def test_CHOOSE(self):
        self.assertEqual(lookup.CHOOSE('2', 2, 4, 6), 4)

    def test_CHOOSE_with_negative_index(self):
        self.assertIsInstance(
            lookup.CHOOSE(-1, 1, 2, 3), xlerrors.ValueExcelError)

    def test_CHOOSE_with_too_large_index(self):
        self.assertIsInstance(
            lookup.CHOOSE(5, 1, 2, 3), xlerrors.ValueExcelError)

    def test_VLOOOKUP(self):
        # Excel Doc example.
        range1 = func_xltypes.Array([
            [101, 'Davis', 'Sara'],
            [102, 'Fortana', 'Olivier'],
            [103, 'Leal', 'Karina'],
            [104, 'Patten', 'Michael'],
            [105, 'Burke', 'Brian'],
            [106, 'Sousa', 'Luis'],
        ])
        self.assertEqual(lookup.VLOOKUP(102, range1, 2, False), 'Fortana')

    def test_VLOOOKUP_with_range_lookup(self):
        with self.assertRaises(NotImplementedError):
            lookup.VLOOKUP(1, func_xltypes.Array([[]]), 2, True)

    def test_VLOOOKUP_with_oversized_col_index_num(self):
        # Excel Doc example.
        range1 = func_xltypes.Array([
            [101, 'Davis', 'Sara'],
        ])
        self.assertIsInstance(
            lookup.VLOOKUP(102, range1, 4, False), xlerrors.ValueExcelError)

    def test_VLOOOKUP_with_unknown_lookup_value(self):
        range1 = func_xltypes.Array([
            [101, 'Davis', 'Sara'],
        ])
        self.assertIsInstance(
            lookup.VLOOKUP(102, range1, 2, False), xlerrors.NaExcelError)

    def test_MATCH(self):
        range1 = [25, 28, 40, 41]
        self.assertEqual(lookup.MATCH(39, range1), 2)
        self.assertEqual(lookup.MATCH(39, range1, 1), 2)
        self.assertEqual(lookup.MATCH(40, range1, 1), 3)
        self.assertEqual(lookup.MATCH(40, range1, 0), 3)
        self.assertIsInstance(
            lookup.MATCH(40, range1, -1), xlerrors.NaExcelError
        )
        range2 = list(reversed(range1))
        self.assertEqual(lookup.MATCH(40, range2, -1), 2)
        self.assertIsInstance(
            lookup.MATCH(100, range2, -1), xlerrors.NaExcelError
        )
        self.assertIsInstance(
            lookup.MATCH(40, range2, 1), xlerrors.NaExcelError
        )
        self.assertIsInstance(
            lookup.MATCH(0, range1, 0), xlerrors.NaExcelError
        )



import unittest
from xlcalculator.xlfunctions import logical, func_xltypes, xlerrors


VExpr = func_xltypes.ValueExpr


class LogicalModuleTest(unittest.TestCase):

    def test_AND(self):
        self.assertEqual(logical.AND(VExpr(3), VExpr(True), VExpr(None)), True)
        self.assertEqual(logical.AND(VExpr(True), VExpr(0)), False)

    def test_AND_without_any_args(self):
        self.assertIsInstance(logical.AND(), xlerrors.NullExcelError)

    def test_AND_with_direct_values(self):
        self.assertEqual(logical.AND(3, True, None), True)
        self.assertEqual(logical.AND(True, 0), False)

    def test_FALSE(self):
        self.assertFalse(logical.FALSE())

    def test_OR(self):
        self.assertEqual(logical.OR(VExpr(None), VExpr(3)), True)
        self.assertEqual(logical.OR(VExpr(False), VExpr(0)), False)

    def test_OR_without_any_args(self):
        self.assertIsInstance(logical.OR(), xlerrors.NullExcelError)

    def test_OR_with_dirext_values(self):
        self.assertEqual(logical.OR(3, True, None), True)
        self.assertEqual(logical.OR(False, 0), False)

    def test_IF(self):
        self.assertEqual(logical.IF(VExpr(True), VExpr(1), VExpr(2)), 1)
        self.assertEqual(logical.IF(VExpr(False), VExpr(1), VExpr(2)), 2)

    def test_IF_with_direct_values(self):
        # Values get converted to expressions to not break the APIs.
        self.assertEqual(logical.IF(True, 1, 2), 1)
        self.assertEqual(logical.IF(False, 1, 2), 2)

    def test_TRUE(self):
        self.assertTrue(logical.TRUE())



from .. import testing


class COSTest(testing.FunctionalTestCase):

    filename = "COS.xlsx"

    def test_evaluation_A2(self):
        self.assertEqual(
            self.evaluator.evaluate('Sheet1!A2'),
            self.evaluator.get_cell_value('Sheet1!A2')
        )

    def test_evaluation_A3(self):
        self.assertEqual(
            self.evaluator.evaluate('Sheet1!A3'),
            self.evaluator.get_cell_value('Sheet1!A3')
        )

    def test_evaluation_A4(self):
        self.assertEqual(
            self.evaluator.evaluate('Sheet1!A4'),
            self.evaluator.get_cell_value('Sheet1!A4')
        )




from decimal import Decimal
from decimal import ROUND_FLOOR
from math import e


def floating_point_rounded():

    print("e", e * 10**-17)
    precision_mask = "{0:." + str(17) + "f}1"
    precision = precision_mask.format(e * 10**-17)


    the_float_lhs = Decimal( float(0.5805555555555556) )
    the_dec_lhs = Decimal('0.5805555555555556')
    lhs_result = Decimal(the_float_lhs).quantize(Decimal(precision), rounding=ROUND_FLOOR)
    print("rounded_float", lhs_result)




    the_float_rhs = Decimal( float(0.5805555555555) )
    the_dec_rhs = Decimal('0.5805555555555')
    rhs_result = Decimal(the_float_rhs).quantize(Decimal(precision), rounding=ROUND_FLOOR)
    print("rounded_float", rhs_result )

    print("             ", precision)
    print("rounded_float",str(lhs_result))
    print("rounded_float", round(rhs_result, 13))
    print("LHS error ", the_float_lhs - the_dec_lhs)
    print("RHS error", the_float_rhs - the_dec_rhs)



if __name__ == '__main__':
    floating_point_rounded()
